# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np

import gwbench.utils as utils
from gwbench.basic_relations import tau_spa_PN35
from gwbench.utils import REarth, time_to_rad_earth, cLight, log_msg, mod_1

cos = np.cos
sin = np.sin
exp = np.exp
log = np.log
pi  = np.pi

ap_symbs_string = 'f Mc eta tc ra dec psi'

def detector_response(hfp, hfc, f, Mc, eta, tc, ra, dec, psi, loc, use_rot, user_locs=None):
    '''
    Calculate the detector response for a given detector location and orientation.

    Parameters
    ----------
    hfp : np.ndarray
        The plus polarization.
    hfc : np.ndarray
        The cross polarization.
    f : np.ndarray
        The frequency domain array [Hz].
    Mc : float
        The chirp Mass [solar mass].
    eta: float
        The symmetric mass ratio.
    tc : float
        The time of coalescence [s].
    dec : float
        The declination [rad].
    ra : float
        The right ascension [rad].
    psi : float
        The polarization angle [rad].
    loc : str
        The location (and implied orientation) of a detector.
    use_rot : bool
        Use frequency dependent time due to rotation of earth and SPA.
    user_locs : dict, optional
        User defined locations and orientations of detectors.

    Returns
    -------
    hf : np.ndarray
        The detector response in the frequency domain.
    '''
    Fp, Fc, Flp = antenna_pattern_and_loc_phase_fac(f, Mc, eta, tc, ra, dec, psi, loc, use_rot, user_locs=user_locs)
    return Flp * (Fp * hfp + Fc * hfc)

def antenna_pattern_and_loc_phase_fac(f, Mc, eta, tc, ra, dec, psi, loc, use_rot, user_locs=None):
    '''
    Calculate the antenna pattern and location phase factor for a given detector location and orientation.

    Parameters
    ----------
    f : np.ndarray
        The frequency domain [Hz].
    Mc : float
        The chirp Mass [solar mass].
    eta: float
        The symmetric mass ratio.
    tc : float
        The time of coalescence [s].
    dec : float
        The declination [rad].
    ra : float
        The right ascencsion [rad].
    psi : float
        The polarization angle [rad].
    loc : str
        The location (and implied orientation) of a detector.
    use_rot : bool
        Use frequency dependent time due to rotation of earth and SPA.
    user_locs : dict, optional
        User defined locations and orientations of detectors.

    Returns
    -------
    Fp : np.ndarray
        The plus polarization antenna pattern.
    Fc : np.ndarray
        The cross polarization antenna pattern.
    Flp : np.ndarray
        The location phase factor.
    '''
    D, d       = det_ten_and_loc_vec(loc, REarth, user_locs=user_locs)
    time       = calc_rotating_time(tc, f, Mc, eta, use_rot)
    time_delay = calc_time_delay(calc_gra(ra, time), dec, d)

    return *ant_pat_funcs(D, *ant_pat_vectors(calc_gra(ra, time + time_delay), dec, psi)), loc_phase_func(f, time_delay)

def calc_rotating_time(tc, f, Mc, eta, use_rot):
    '''
    Calculate the time including the effects of the rotation of the earth
    using the stationary phase approximation, if use_rot is True.

    Parameters
    ----------
    tc : float
        The time of coalescence [s].
    f : np.ndarray
        The frequency domain [Hz].
    Mc : float
        The chirp Mass [solar mass
    use_rot : bool
        Include rotation effects.

    Returns
    -------
    time : np.ndarray
        The time including rotation effects [s].
    '''
    if use_rot: return tc + utils.tc_offset - tau_spa_PN35(f, Mc, eta, log=log)
    else:       return np.array([tc + utils.tc_offset])

def calc_gra(ra, time):
    '''
    Calculate the Greenwich Right Ascension (GRA) for a given detector location and orientation.

    Parameters
    ----------
    ra : float
        The right ascencsion [rad].
    time : np.ndarray
        The corrected time [s].

    Returns
    -------
    gra : np.ndarray
        The Greenwich Right Ascension [rad].
    '''
    return ra - time_to_rad_earth * time

def calc_time_delay(gra, dec, d):
    '''
    Calculate the time delay from the geocenter for a given detector location and orientation.

    Parameters
    ----------
    gra : np.ndarray
        Greenwich Right Ascension [rad].
    dec : float
        Declination [rad].
    d : np.ndarray
        Detector location vector.

    Returns
    -------
    time_delay : np.ndarray
        Time delay from the geocenter [s].
    '''
    # using cos/sin(dec) instead of cos/sin(theta) with polar angle (theta = pi/2 - dec)
    return np.matmul(d, np.array([cos(gra)*cos(dec), sin(gra)*cos(dec), sin(dec)*np.ones_like(gra)]))

def loc_phase_func(f, time_delay):
    '''
    Calculate the location phase factor encoding the phase difference between the signal at the
    detector and the signal at the geocenter.

    Parameters
    ----------
    f : np.ndarray
        Frequency domain [Hz].
    time_delay : np.ndarray
        The corrected time [s].

    Returns
    -------
    Flp : np.ndarray
        Location phase factor
    '''
    return exp(1j * 2*pi * mod_1(f * time_delay, np=np))

def ant_pat_funcs(D, XX, YY):
    '''
    Calculate the antenna pattern for a given detector location and orientation.

    Parameters
    ----------
    D : np.ndarray
        The detector tensor.
    XX : np.ndarray
        The x-arm antenna pattern vector.
    YY : np.ndarray
        The y-arm antenna pattern vector.

    Returns
    -------
    Fp : np.ndarray
        The plus polarization antenna pattern.
    Fc : np.ndarray
        The cross polarization antenna pattern.
    '''
    return (0.5 * (np.matmul(D,XX) * XX - np.matmul(D,YY) * YY)).sum(axis=0), \
           (0.5 * (np.matmul(D,XX) * YY + np.matmul(D,YY) * XX)).sum(axis=0)

def ant_pat_vectors(gra, dec, psi):
    '''
    Calculate the antenna pattern vectors for a given detector location and orientation.

    Parameters
    ----------
    gra : np.ndarray
        Greenwich Right Ascension [rad]
    dec : float
        Declination [rad]
    psi : float
        Polarization angle [rad]

    Returns
    -------
    XX : np.ndarray
        x-arm antenna pattern vector
    YY : np.ndarray
        y-arm antenna pattern vector
    '''
    return np.array([  cos(psi)*sin(gra) - sin(psi)*cos(gra)*sin(dec),
                      -cos(psi)*cos(gra) - sin(psi)*sin(gra)*sin(dec),
                                np.ones_like(gra) * sin(psi)*cos(dec) ]), \
           np.array([ -sin(psi)*sin(gra) - cos(psi)*cos(gra)*sin(dec),
                       sin(psi)*cos(gra) - cos(psi)*sin(gra)*sin(dec),
                                np.ones_like(gra) * cos(psi)*cos(dec) ])

def det_ten_and_loc_vec(loc, R, user_locs=None):
    '''
    Calculate the detector tensor and location vector for a given detector location and orientation.

    Parameters
    ----------
    loc : str
        Location (and implied orientation) of a detector.
    R : float
        Radius of earth [m].
    user_locs : dict, optional
        User defined locations and orientations of detectors.

    Returns
    -------
    D : np.ndarray
        Detector tensor.
    d : np.ndarray
        Detector location vector.
    '''
    i_vec = np.array([1,0,0])
    j_vec = np.array([0,1,0])
    k_vec = np.array([0,0,1])

    et_vec2 = ( i_vec + np.sqrt(3.)*j_vec)/2.
    et_vec3 = (-i_vec + np.sqrt(3.)*j_vec)/2.

    alpha, beta, gamma, shape = det_angles_shape(loc, user_locs=user_locs)
    # insert polar angle theta = pi/2 - beta instead of latitude beta
    EulerD1 = np.matmul(np.matmul(rot_mat(alpha, 2), rot_mat(pi/2 - beta, 1)), rot_mat(gamma, 2))

    if   shape == 'V3':
        eDArm1 = -1 * np.matmul(EulerD1,et_vec2)
        eDArm2 = -1 * np.matmul(EulerD1,et_vec3)
    elif shape == 'V2':
        eDArm1 =      np.matmul(EulerD1,et_vec3)
        eDArm2 = -1 * np.matmul(EulerD1,i_vec)
    elif shape == 'V1':
        eDArm1 =      np.matmul(EulerD1,i_vec)
        eDArm2 =      np.matmul(EulerD1,et_vec2)
    elif shape == 'L':
        eDArm1 =      np.matmul(EulerD1,i_vec)
        eDArm2 =      np.matmul(EulerD1,j_vec)

    return np.outer(eDArm1,eDArm1) - np.outer(eDArm2,eDArm2), -R/cLight * np.matmul(EulerD1,k_vec)

def rot_mat(angle, axis):
    '''
    Calculate the rotation matrix for a given angle and axis.

    Parameters
    ----------
    angle : float
        Rotation angle [rad]
    axis : int
        Rotation axis (0, 1, or 2)

    Returns
    -------
    rot : np.ndarray
        Rotation matrix
    '''
    c = np.cos(angle)
    s = np.sin(angle)

    if axis == 0: return np.array( [ [1,0,0], [0,c,-s], [0,s,c] ] )
    if axis == 1: return np.array( [ [c,0,s], [0,1,0], [-s,0,c] ] )
    if axis == 2: return np.array( [ [c,-s,0], [s,c,0], [0,0,1] ] )

def correct_arm_azimuth(arm_azimuth, which_arm):
    '''
    Correct the azimuth of the interferometer arm.

    Parameters
    ----------
    arm_azimuth : float
        Azimuth of the interferometer arm [rad]
    which_arm : str
        Interferometer arm ('x', 'y', or 'bisector')

    Returns
    -------
    corr_arm_azimuth : float
        Corrected azimuth of the interferometer arm [rad]
    '''
    if   which_arm == 'x':        return pi/2 + arm_azimuth
    elif which_arm == 'y':        return        arm_azimuth
    elif which_arm == 'bisector': return pi/4 + arm_azimuth
    else: log_msg(f'correct_arm_azimuth: Unknown interferometer arm specifier {which_arm}, available options are: "x", "y", or "bisect"!', level='ERROR')

def det_angles_shape(loc, user_locs=None):
    '''
    Calculate the detector angles and shape for a given detector location.

    Parameters
    ----------
    loc : str
        Location (and implied orientation) of a detector.
    user_locs : dict
        User defined locations and orientations of detectors.

    Returns
    -------
    alpha : float
        Longitude [rad]
    beta : float
        Latitude [rad]
    gamma : float
        Angle from 'Due East' to y-arm [rad] (for in-built locations)
    shape : str
        Shape of interferometer (e.g. 'L', 'V1', ...)

    Note
    ----
    Switched to more precise angles for H, L, V, K, I, ET1, ET2, ET3 on 2022_03_29, found at:
    https://lscsoft.docs.ligo.org/lalsuite/lal/_l_a_l_detectors_8h_source.html
    '''
    if user_locs is not None and loc in user_locs:
        user_loc = user_locs[loc]
        return user_loc['longitude'], user_loc['latitude'], correct_arm_azimuth(user_loc['arm_azimuth'], user_loc['which_arm']), user_loc['shape']
    # 2G sites
    elif loc == 'H':     return -2.08405676917,  0.81079526383, correct_arm_azimuth(pi-5.65487724844, 'y'), 'L'
    elif loc == 'L':     return -1.58430937078,  0.53342313506, correct_arm_azimuth(pi-4.40317772346, 'y'), 'L'
    elif loc == 'V':     return  0.18333805213,  0.76151183984, correct_arm_azimuth(pi-0.33916285222, 'y'), 'L'
    elif loc == 'K':     return  2.396441015,    0.6355068497,  correct_arm_azimuth(pi-1.054113,      'y'), 'L'
    elif loc == 'I':     return  1.33401332494,  0.24841853020, correct_arm_azimuth(pi-1.57079637051, 'y'), 'L'
    elif loc == 'LHO':   return -2.06982474503,  0.81079270401, correct_arm_azimuth(2.19910516123923, 'x'), 'L'
    elif loc == 'LLO':   return -1.5572845695,   0.53342110078, correct_arm_azimuth(3.45080197126464, 'x'), 'L'
    elif loc == 'LIO':   return  1.3444416672,   0.34231239582, correct_arm_azimuth(2.05248780779809, 'x'), 'L'
    # fiducial ET sites
    elif loc == 'ET1':   return  0.18333805213,  0.76151183984, correct_arm_azimuth(pi-0.33916285222, 'y'), 'V1'
    elif loc == 'ET2':   return  0.18333805213,  0.76151183984, correct_arm_azimuth(pi-0.33916285222, 'y'), 'V2'
    elif loc == 'ET3':   return  0.18333805213,  0.76151183984, correct_arm_azimuth(pi-0.33916285222, 'y'), 'V3'
    elif loc == 'ETS1':  return  0.1643518379,   0.70714923527, correct_arm_azimuth(1.5707963267949,  'x'), 'V1'
    elif loc == 'ETS2':  return  0.1643518379,   0.70714923527, correct_arm_azimuth(1.5707963267949,  'x'), 'V2'
    elif loc == 'ETS3':  return  0.1643518379,   0.70714923527, correct_arm_azimuth(1.5707963267949,  'x'), 'V3'
    # fiducial CE sites
    elif loc == 'C':     return -1.9691740,      0.764918,      correct_arm_azimuth(0.,               'y'), 'L'
    elif loc == 'N':     return -1.8584265,      0.578751,      correct_arm_azimuth(-pi/3.,           'y'), 'L'
    elif loc == 'S':     return  2.5307270,     -0.593412,      correct_arm_azimuth(pi/4.,            'y'), 'L'
    elif loc == 'CEA':   return -2.18166156499,  0.80285145592, correct_arm_azimuth(4.53785605518526, 'x'), 'L'
    elif loc == 'CEB':   return -1.64060949687,  0.50614548308, correct_arm_azimuth(3.49065850398866, 'x'), 'L'
    elif loc == 'CES':   return  2.5307270,     -0.593412,      correct_arm_azimuth(pi/4.,            'y'), 'L'
    else: log_msg(f'det_angles_shape: Provided location {loc} is neither among the implemented locations nor in provided user_locs.', level='ERROR')



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""This script calculates the lamdified derivatives of the waveform polarizations, antenna patterns, and detector responses.

The first lines of output are needed for the main benchmarking methods and should be copied to the main script.

"""

import os

from gwbench import antenna_pattern_np as ant_pat_np
from gwbench import detector_response_derivatives as drd
from gwbench import waveform as wfc

############################################################################
### User Choices
############################################################################

#-----choose waveform model - see below for available waveforms-----
# uncomment chosen waveform model from the list of currently available
if 1:
    wf_model_name = 'tf2'
    #wf_model_name = 'tf2_tidal'

    user_waveform = None

    #-----choose partial derivative variables for the chose waveform model-----
    if wf_model_name == 'tf2':
        wf_other_var_dic = None
        deriv_symbs_string = 'Mc eta chi1z chi2z DL tc phic iota ra dec psi'

    elif wf_model_name == 'tf2_tidal':
        wf_other_var_dic = None
        deriv_symbs_string = 'Mc eta chi1z chi2z DL tc phic iota lam_t ra dec psi'

else:
    wf_model_name    = 'tf2_user'
    wf_other_var_dic = None
    user_waveform   = {'np': '../gwbench/wf_models/tf2_np.py', 'sp':'../gwbench/wf_models/tf2_sp.py'}

#-----choose locations whose detector response derivatives should be calculated-----
# pass a list or None (if all locations, available in gwbench, are supposed to be used)
locs = ['H', 'L', 'V', 'K', 'I', 'ET1', 'ET2', 'ET3', 'C', 'N', 'S', 'user-loc']

#-----choose whether antenna patterns and location phase factors should use frquency dependent gra, thus incorporating the rotation of earth-----
# 1 for True, 0 for False - use 0 for speed or when comparability to code not incorporating the rotation of earth
use_rot = 1

#-----choose where to save the lambidified functions files-----
output_path = None

#-----user defined locations-----
user_locs = {'user-loc':{'longitude': 3.2, 'latitude': 0.4, 'arm_azimuth':0.3, 'which_arm':'y', 'shape':'L'}}

############################################################################
### Calculation of Lambdified Functions
############################################################################

drd.generate_det_responses_derivs_sym(wf_model_name, wf_other_var_dic, deriv_symbs_string, use_rot,
                                      locs=locs, user_locs=user_locs, user_waveform=user_waveform,
                                      user_lambdified_functions_path=output_path)



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np
from gwbench import network

############################################################################
### User Choices
############################################################################

# include earth's rotation for Fp, Fc, lpf (location phase factor)
# (if set = 1, inj_params['Mc'] and inj_params['tc'] cannot be None)
use_rot = 0

# injection parameters (Mc, tc are only needed, if the motion of earth is relevant)
inj_params = {
    'tc':    0.,
    'ra':    np.pi/4,
    'dec':   np.pi/4,
    'psi':   np.pi/4,
    }

# choose frequency array
f_lo = 1.
f_hi = 2.**12
df = 2.**-4
f = np.arange(f_lo,f_hi+df,df)

# choose detectors/network
network_label = 'ECa4cSa4c'

############################################################################
### Initialize network and setup antenna patterns, location phase factors,
### and PSDs
############################################################################

net = network.Network(network_label)
net.set_net_vars(f=f,inj_params=inj_params,use_rot=use_rot)
net.setup_ant_pat_lpf_psds()

############################################################################
### Get antenna patterns, location phase factors, and PSDs
############################################################################

print()

# Method 1 - using the detector keys to get specific detectors
print('Method 1')
print(net.det_keys)
print()
for det_key in net.det_keys:
    print(det_key)
    det = net.get_detector(det_key)
    print(det.f)
    print(det.Fp)
    print(det.Fc)
    print(det.Flp)
    print(det.psd)
    print()

print()

# Method 2 - cycling through the detectors directly
print('Method 2')
print(net.detectors)
print()
for det in net.detectors:
    print(det.det_key)
    print(det.f)
    print(det.Fp)
    print(det.Fc)
    print(det.Flp)
    print(det.psd)
    print()



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import numpy as np
from gwbench import Network, M_of_Mc_eta, f_isco_Msolar

#import gwbench
#gwbench.utils.tc_offset = 0

############################################################################
### settings
############################################################################

# choose which parameters to take partial derivatives for Fisher analysis
deriv_symbs_string  = 'Mc eta chi1z chi2z DL tc phic iota ra dec psi'
# choose which parameters to convert to cos (None) or log versions
conv_log            = ('Mc','DL')
# choose whether to take Earth's rotation into account
use_rot             = True
# choose whether to perform the Fisher analysis only at the network level
only_net            = True

############################################################################
### injection parameters and frequency array
############################################################################

# set the injection parameters
inj_params = {
    'Mc'    : 42.92123785996995,
    'eta'   : 0.2179560548922493,
    'chi1x' : 0.0,
    'chi1y' : 0.0,
    'chi1z' : -0.5291825652680437,
    'chi2x' : 0.0,
    'chi2y' : 0.0,
    'chi2z' : -0.5533273077110299,
    'DL'    : 595.845227743203,
    'tc'    : 0.0,
    'phic'  : 0.0,
    'iota'  : 0.2367522187410996,
    'ra'    : 4.043179805492967,
    'dec'   : 1.1829476432948405,
    'psi'   : 3.277322349164817,
    }

print('injections parameter: ', inj_params)
print()

# pick the desired frequency range
f_lo = 1.
f_hi = f_isco_Msolar(M_of_Mc_eta(inj_params['Mc'],inj_params['eta']))
df   = 2.**-4
f    = np.arange(f_lo,f_hi+df,df)

print('f_lo:', f_lo, '   f_hi:', f_hi, '   df:', df)
print()

############################################################################
### Network specification
############################################################################

# choose the desired detectors
network_spec = ['CE2-40-CBO_C','CE2-40-CBO_N','CE2-40-CBO_S']
print('network spec: ', network_spec)
print()

############################################################################
### GW Benchmarking
############################################################################

# initialize the network with the desired detectors
nnet = Network(network_spec, logger_name='3_40km_CE', logger_level='INFO')
# pass all variables to the network
nnet.set_net_vars(wf_model_name='tf2', f=f, inj_params=inj_params,
                  deriv_symbs_string=deriv_symbs_string, conv_log=conv_log, use_rot=use_rot)

# calculate the network and detector Fisher matrices, condition numbers,
# covariance matrices, error estimates (including 90%-credible sky area
# in deg), and inversion errors using numeric differentiation mehtod
nnet.calc_errors(only_net=only_net, derivs='num', step=1e-9, method='central', order=2)

############################################################################
### Check results
############################################################################

# stored from previous evaluation for tf2 waveform and inj_id=0
snr  = 2598
errs = {
    'log_Mc'      : 0.001872,
    'eta'         : 0.078637,
    'chi1z'       : 1.355901,
    'chi2z'       : 2.016472,
    'log_DL'      : 0.036482,
    'tc'          : 0.005800,
    'phic'        : 16.09812,
    'iota'        : 0.150074,
    'ra'          : 0.001220,
    'dec'         : 0.000657,
    'psi'         : 0.666370,
    'sky_area_90' : 0.011030,
    }

rtol = 1e-2
atol = 0
print(f'\nCheck if calculated and stored values agree up to a relative error of {rtol}.')
print(f'{"Network SNR".ljust(19)} calculated={str(nnet.snr).ljust(20)}   stored={str(snr).ljust(18)} agree={np.isclose(nnet.snr, snr, atol=atol, rtol=rtol)}.')
for key in errs:
    cval = nnet.errs[key]
    sval = errs[key]
    print(f'Error {key.ljust(13)} calculated={str(cval).ljust(22)} stored={str(sval).ljust(18)} agree={np.isclose(cval, sval, atol=atol, rtol=rtol)}.')
print()

############################################################################
### Check against jax method
############################################################################

# initialize the network with the desired detectors
jnet = Network(network_spec, logger_name='3_40km_CE', logger_level='INFO')
# pass all variables to the network
jnet.set_net_vars(wf_model_name='tf2_jx', f=f, inj_params=inj_params,
                  deriv_symbs_string=deriv_symbs_string, conv_log=conv_log, use_rot=use_rot)

# calculate the network and detector Fisher matrices, condition numbers,
# covariance matrices, error estimates (including 90%-credible sky area
# in deg), and inversion errors using numeric jax differentiation mehtod
jnet.calc_errors(only_net=only_net, derivs='num')

rtol = 5e-2
print(f'\nCheck if numericly calculated (numdifftools vs jax) errors agree up to a relative error of {rtol}.')
for key in errs:
    sval = nnet.errs[key]
    cval = jnet.errs[key]
    print(f'Error {key.ljust(13)}    numdifftools={str(sval).ljust(22)} jax={str(cval).ljust(22)} agree={np.isclose(cval, sval, atol=atol, rtol=rtol)}.')
print()

############################################################################
### Check against symbolic method
############################################################################

# initialize the network with the desired detectors
snet = Network(network_spec, logger_name='3_40km_CE', logger_level='INFO')
# pass all variables to the network
snet.set_net_vars(wf_model_name='tf2', f=f, inj_params=inj_params,
                  deriv_symbs_string=deriv_symbs_string, conv_log=conv_log, use_rot=use_rot)

# calculate the network and detector Fisher matrices, condition numbers,
# covariance matrices, error estimates (including 90%-credible sky area
# in deg), and inversion errors using symbolic differentiation mehtod
snet.calc_errors(only_net=only_net, derivs='sym', gen_derivs=True)

rtol = 5e-2
print(f'\nCheck if numericly and symbolicly calculated errors agree up to a relative error of {rtol}.')
for key in errs:
    sval = nnet.errs[key]
    cval = snet.errs[key]
    print(f'Error {key.ljust(13)}    numeric={str(sval).ljust(22)} symbolic={str(cval).ljust(22)} agree={np.isclose(cval, sval, atol=atol, rtol=rtol)}.')



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import sympy as sp

from gwbench.wf_models.tf2_tidal import hfpc_raw, wf_symbs_string
from gwbench.wf_models.tf2_tidal_np import hfpc

deriv_mod = 'numdifftools'

f, Mc, eta, chi1z, chi2z, DL, tc, phic, iota, lam_t, delta_lam_t = sp.symbols(wf_symbs_string, real=True)


def hfpc_sp(f, Mc, eta, chi1z, chi2z, DL, tc, phic, iota, lam_t, delta_lam_t, is_lam12=False):
     '''
     Sympy implementation of TaylorF2 waveform model for BNS systems.
     For further details check gwbench.wf_models.tf2_tidal.hfpc
     '''
     return hfpc_raw(f, Mc, eta, chi1z, chi2z, DL, tc, phic, iota, lam_t, delta_lam_t, sp.cos, sp.sin, sp.exp, sp.log, is_lam12=is_lam12)



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np
from scipy.integrate import simpson
from scipy.optimize import toms748

#-----derivative and error manipulations-----
def convert_to_cos_derivative(deriv, param_key, param_val):
    '''
    Convert the partial derivative [del_X(f)] of a function [f] with respect to a parameter [X]
    to the derivative [del_cos_X(f)] with respect to the cosine of the parameter.

    Parameters
    ----------
    deriv : float
        the partial derivative [del_X(f)] of the function [f] with respect to the parameter [X].
    param_key : str
        the name of the parameter [X].
    param_val : float
        the value of the parameter [X].

    Returns
    -------
    c_deriv : float
        the derivative [del_cos_X(f)] with respect to the cosine of the parameter.
    c_param_key : str
        the name of the converted parameter [cos_X].
    c_param_val : float
        the value of the parameter [cos_X].
    '''
    c_deriv     = (-1./np.sin(param_val)) * deriv
    c_param_key = 'cos_'+param_key
    c_param_val = np.cos(param_val)
    return c_deriv, c_param_key, c_param_val

def convert_to_log_derivative(deriv, param_key, param_val):
    '''
    Convert the partial derivative [del_X(f)] of a function [f] with respect to a parameter [X]
    to the derivative [del_log_X(f)] with respect to the logarithm of the parameter.

    Parameters
    ----------
    deriv : float
        the partial derivative [del_X(f)] of the function [f] with respect to the parameter [X].
    param_key : str
        the name of the parameter [X].
    param_val : float
        the value of the parameter [X].

    Returns
    -------
    c_deriv : float
        the derivative [del_log_X(f)] with respect to the logarithm of the parameter.
    c_param_key : str
        the name of the converted parameter [log_X].
    c_param_val : float
        the value of the parameter [log_X].
    '''
    c_deriv     = param_val * deriv
    c_param_key = 'log_'+param_key
    c_param_val = np.log(param_val)
    return c_deriv, c_param_key, c_param_val

def dim_err_to_rel_err(err, param_val, param_key=None, param_kind=None):
    '''
    Convert the error [err] of a dimensional parameter [X] to the relative error [err_rel] with respect to the parameter [X].

    Parameters
    ----------
    err : float
        the error of the parameter [X].
    param_val : float
        the value of the parameter [X].
    param_key : str, optional
        the name of the parameter [X].
    param_kind : str, optional
        the kind of the parameter [X].

    Returns
    -------
    err_rel : float
        the relative error of the parameter [X].
    '''
    if param_key in ('M','Mc','Dl','DL','tc') or param_kind == 'dim': return np.abs(err/param_val)
    else:                                                             return err

def one_sigma_to_percent_error(percent, sigma):
    '''
    Convert the one-sigma error [sigma] to the percent error [percent].

    Parameters
    ----------
    percent : float
        the percentage desired as a fraction.
    sigma : float
        the one-sigma error.

    Returns
    -------
    percent : float
        the percent error.
    '''
    sigma_orders = [0, 1, 2, 3, 4, 5, 6]
    sigma_percents = [0., 68.2689492137, 95.4499736104, 99.7300203937, 99.9936657516, 99.9999426697, 99.9999998027]
    for i,sigma_order in enumerate(sigma_orders):
        if percent < sigma_percents[i]: break

    def func(error,percent,sigma,sigma_order):
        x = np.linspace(0,error,sigma_order*100)
        return percent/100/2 - simpson(np.exp(-x**2/2/sigma**2)/np.sqrt(2*np.pi)/sigma, x=x)

    if sigma_order == 1: return toms748(func,1e-2*sigma,sigma_order*sigma,args=(percent,sigma,sigma_order))
    else:                return toms748(func,(sigma_order-1)*sigma,sigma_order*sigma,args=(percent,sigma,sigma_order))

#-----Convert the evaluated derivatives according to conv_cos, conv_log-----
def get_conv_inj_params_deriv_variables(c_quants, inj_params, deriv_variables):
    '''
    Convert the injected parameters and the differentiated variables string according to the conversion dictionary [c_quants].

    Parameters
    ----------
    c_quants : dict
        the conversion dictionary.
    inj_params : dict
        the injected parameters.
    deriv_variables : list
        the differentiation variables.

    Returns
    -------
    inj_params : dict
        the converted injected parameters.
    deriv_variables : list
        the converted differentiation variables.
    '''
    if c_quants == {}: return inj_params, deriv_variables
    else:
        for o_key in c_quants:
            c_key = c_quants[o_key][0]
            c_val = c_quants[o_key][1]

            if c_key not in deriv_variables: deriv_variables[deriv_variables.index(o_key)] = c_key
            inj_params[c_key] = c_val

        return inj_params, deriv_variables

def get_conv_del_eval_dic(del_eval_dic, params_dic, conv_cos, conv_log):
    '''
    Convert the evaluated derivatives according to the conversion dictionaries [conv_cos, conv_log].

    Parameters
    ----------
    del_eval_dic : dict
        the evaluated derivatives.
    params_dic : dict
        the parameters dictionary.
    conv_cos : dict
        the conversion dictionary for the cosine of the parameters.
    conv_log : dict
        the conversion dictionary for the logarithm of the parameters.
    deriv_symbs_string : str
        the string of the differentiation variables.

    Returns
    -------
    conv_dic : dict
        the converted evaluated derivatives.
    c_quants : dict
        the converted parameters and their values.
    '''
    if conv_cos is None and conv_log is None: return del_eval_dic, {}
    else:
        conv_dic = {}
        c_quants = {}

        for deriv in del_eval_dic:
            key = '_'.join(deriv.split('_')[1:-1])
            c_key = None

            if   conv_cos is not None and key in conv_cos: c_deriv, c_key, c_val = convert_to_cos_derivative(del_eval_dic[deriv],key,params_dic[key])
            elif conv_log is not None and key in conv_log: c_deriv, c_key, c_val = convert_to_log_derivative(del_eval_dic[deriv],key,params_dic[key])
            else:                                          conv_dic[deriv]       = del_eval_dic[deriv]

            if c_key is not None:
                c_quants[key]   = (c_key, c_val)
                prefix,suffix   = deriv.split(key)
                n_key           = prefix + c_key + suffix
                conv_dic[n_key] = c_deriv

        return conv_dic, c_quants

#-----sky area calculations-----
def sky_area_90(ra_err, dec_err, cov_ra_dec, dec_val, is_cos_dec):
    '''
    Calculate the 90% credible sky area in square degrees.

    Parameters
    ----------
    ra_err : float
        the error of the right ascension in radians.
    dec_err : float
        the error of the declination in radians.
    cov_ra_dec : float
        the covariance between the right ascension and the declination.
    dec_val : float
        the value of the declination in radians.
    is_cos_dec : bool
        if True, the declination is in cosine.

    Returns
    -------
    sky_area : float
        the 90% credible sky area in square degrees.

    References
    ----------
    [1] arXiv:1403.6915
    [2] arXiv:gr-qc/0310125v3
    '''
    if is_cos_dec: trig_fac = np.abs( 1/np.tan(dec_val) )
    else:          trig_fac = np.abs(   np.cos(dec_val) )
    if (ra_err*dec_err)**2 > cov_ra_dec**2: return trig_fac * np.sqrt( (ra_err * dec_err)**2 - cov_ra_dec**2 ) * 2*np.pi * (180./np.pi)**2 * np.log(10)
    else:                                   return None



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import astropy as ap
import numpy as np
from lal import GreenwichMeanSiderealTime

from gwbench.utils import MTsun, time_to_rad_earth

#-----GMST and GPS time conversions-----
def gmst_of_gps_time(gps_time, radians=False):
    '''
    Calculate the Greenwich Mean Sidereal Time (GMST) given the GPS time.
    Wrapper of the `lal.GreenwichMeanSiderealTime` function.

    Parameters
    ----------
    gps_time : float
        The GPS time in seconds.
    radians : bool, optional
        If True, return the GMST in radians, otherwise in seconds. Default is False.

    Returns
    -------
    gmst : float
        The Greenwich Mean Sidereal Time in seconds or radians.
    '''
    if radians: return GreenwichMeanSiderealTime(gps_time)
    else:       return GreenwichMeanSiderealTime(gps_time) / time_to_rad_earth


#-----Keplerian velocity formula-----
def vel_kep(f, M):
    '''
    Calculate the Keplerian velocity for a given frequency and mass.

    Parameters
    ----------
    f : np.ndarray
        The frequency in Hz.
    M : float
        The mass in seconds.

    Returns
    -------
    v : np.ndarray
        The velocity.
    '''
    return (np.pi * M * f)**(1./3)

def vel_kep_Msolar(f, M):
    '''
    Calculate the Keplerian velocity for a given frequency and mass.

    Parameters
    ----------
    f : np.ndarray
        The frequency in Hz.
    M : float
        The mass in solar mass.

    Returns
    -------
    v : np.ndarray
        The velocity.
    '''
    return vel_kep(f, M * MTsun)


#-----tau_spa-----
def tau_spa_PN0(f, Mc):
    '''
    Calculate the frequency-dependent time tau_spa (tf) in the stationary-phase-approximation to account for Earth's rotation.
    see Eq. (7) - (9) in https://arxiv.org/abs/1710.05325v2

    This 0PN correction is taken from the above source, see text after Eq. (9),
    which is taken from Eq. (4.19) from the book "Gravitational Waves: Volume 1: Theory and Experiments" by Michele Maggiore.

    Parameters
    ----------
    f : np.ndarray
        The frequency in Hz.
    Mc : float
        The chirp mass in seconds.

    Returns
    -------
    tf : np.ndarray
        The frequency-dependent time in seconds.
    '''
    return 5. / 256 * (MTsun * Mc)**(-5./3) * (np.pi * f)**(-8./3)

def tau_spa_PN35(f, Mc, eta, log=np.log):
    '''
    Calculate the frequency-dependent time tau_spa (tf) in the stationary-phase-approximation to account for Earth's rotation.
    see Eq. (7) - (9) in https://arxiv.org/abs/1710.05325v2

    This 3.5PN correction is taken from Eq. (3.8b) in https://arxiv.org/abs/0907.0700

    Parameters
    ----------
    f : np.ndarray
        The frequency in Hz.
    Mc : float
        The chirp mass in seconds.

    Returns
    -------
    tf : np.ndarray
        The frequency-dependent time in seconds.
    '''
    eta2 = eta*eta
    pi2  = np.pi*np.pi
    M    = MTsun * M_of_Mc_eta(Mc, eta)
    vel  = vel_kep(f, M)
    vel2 = vel**2
    vel4 = vel2**2
    vel6 = vel2*vel4

    return  5. * M / (256 * eta * vel2*vel6) * \
        (1 +
        (743./252 + 11./3 * eta) * vel2 -
        32 / 5 * np.pi * vel*vel2 +
        (3058673./508032 + 5429./504 * eta + 617./72 * eta2) * vel4 -
        (7729./252 - 13./3 * eta) * np.pi * vel*vel4 +
        (-10052469856691./23471078400 + 128./3 * pi2 + 6848./105 * np.euler_gamma +
        (3147553127./3048192 - 451./12 * pi2) * eta - 15211./1728 * eta2 +
        25565./1296 * eta*eta2 + 3424./105 * log(16. * vel2)) * vel6 +
        (-15419335./127008 - 75703./756 * eta + 14809./378 * eta2) * np.pi * vel*vel6)


#-----f_isco-----
def f_isco(M):
    '''
    Calculate the frequency at the innermost stable circular orbit (ISCO) for a given mass.

    Parameters
    -----------
    M : float
        The mass in seconds.

    Returns
    -------
    f_isco : float
        The frequency at the ISCO.
    '''
    return 1./6.**(3./2.)/np.pi/M

def f_isco_Msolar(M):
    '''
    Calculate the frequency at the innermost stable circular orbit (ISCO) for a given mass.

    Parameters
    -----------
    M : float
        The mass in solar mass.

    Returns
    -------
    f_isco : float
        The frequency at the ISCO.
    '''
    # convert to sec
    return f_isco(M * MTsun)


#-----f_ew (early warning frequency upper cutoff)-----
def f_ew(tau_ew, M, eta):
    '''
    Calculate the early warning frequency upper cutoff for a given mass and symmetric mass ratio.

    Parameters
    -----------
    tau_ew : float
        The early warning time in seconds.
    M : float
        The mass in seconds.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    f_ew : float
        The early warning frequency upper cutoff.
    '''
    return (5 * M / 256 / eta / tau_ew)**(3/8) / np.pi / M

def f_ew_Msolar(tau_ew, M, eta):
    '''
    Calculate the early warning frequency upper cutoff for a given mass and symmetric mass ratio.

    Parameters
    -----------
    tau_ew : float
        The early warning time in seconds.
    M : float
        The mass in solar mass.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    f_ew : float
        The early warning frequency upper cutoff.
    '''
    # convert to sec
    return f_ew(tau_ew, M * MTsun, eta)


#-----mass functions-----
def eta_of_q(q):
    '''
    Calculate the symmetric mass ratio (eta) given the mass ratio (q).

    Parameters
    ----------
    q : float
        The mass ratio.

    Returns
    -------
    eta : float
        The symmetric mass ratio.
    '''
    return q / (1 + q)**2

def delta_of_q(q):
    '''
    Calculate the delta mass difference parameter given the mass ratio (q).

    Parameters
    ----------
    q : float
        The mass ratio.

    Returns
    -------
    delta : float
        The mass difference parameter.
    '''
    return (1 - 4 * eta_of_q(q))**0.5

def delta_of_eta(eta):
    '''
    Calculate the mass difference parameter (delta) given the symmetric mass ratio (eta).

    Parameters
    ----------
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    delta : float
        The mass difference parameter.
    '''
    return (1 - 4 * eta)**0.5

def q_of_eta(eta, q_gt_1=True):
    '''
    Calculate the mass ratio (q) given the symmetric mass ratio (eta).

    Parameters
    ----------
    eta : float
        The symmetric mass ratio.
    q_gt_1 : bool, optional
        The mass ratio q is greater than 1. Default is True.

    Returns
    -------
    q : float
        The mass ratio.
    '''
    if q_gt_1: return (1 + delta_of_eta(eta)) / (1 - delta_of_eta(eta))
    else:      return (1 - delta_of_eta(eta)) / (1 + delta_of_eta(eta))

def q_of_m1_m2(m1, m2, q_gt_1=1):
    '''
    Calculate the mass ratio (q) given the component masses (m1, m2).

    Parameters
    ----------
    m1 : float
        Mass of the first binary component.
    m2 : float
        Mass of the second binary component.
    q_gt_1 : bool, optional
        The mass ratio q is greater than 1. Default is True.

    Returns
    -------
    q : float
        The mass ratio.
    '''
    if q_gt_1: return m1 / m2
    else:      return m2 / m1

def M_of_Mc_eta(Mc, eta):
    '''
    Calculate the total mass (M) given the chirp mass (Mc) and symmetric mass ratio (eta).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    M : float
        The total mass.
    '''
    return Mc / eta**0.6

def M_of_Mc_q(Mc, q):
    '''
    Calculate the total mass (M) given the chirp mass (Mc) and mass ratio (q).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    q : float
        The mass ratio.

    Returns
    -------
    M : float
        The total mass.
    '''
    return Mc / eta_of_q(q)**0.6

def Mc_of_M_eta(M, eta):
    '''
    Calculate the chirp mass (Mc) given the total mass (M) and symmetric mass ratio (eta).

    Parameters
    ----------
    M : float
        The total mass.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    Mc : float
        The chirp mass.
    '''
    return M * eta**0.6

def Mc_of_M_q(M, q):
    '''
    Calculate the chirp mass (Mc) given the total mass (M) and mass ratio (q).

    Parameters
    ----------
    M : float
        The total mass.
    q : float
        The mass ratio.

    Returns
    -------
    Mc : float
        The chirp mass.
    '''
    return M * eta_of_q(q)**0.6

def Mc_of_m1_m2(m1, m2):
    '''
    Calculate the chirp mass (Mc) given the component masses (m1, m2).

    Parameters
    ----------
    m1 : float
        Mass of the first component.
    m2 : float
        Mass of the second component.

    Returns
    -------
    Mc : float
        The chirp mass.
    '''
    return Mc_of_M_eta(m1 + m2, eta_of_q(m1 / m2))

def m1_of_M_eta(M, eta):
    '''
    Calculate the mass of the primary object (m1) in a binary system
    given the total mass (M) and the symmetric mass ratio (eta).

    Parameters
    ----------
    M : float
        Total mass of the binary system.
    eta : float
        Symmetric mass ratio of the binary system.

    Returns
    -------
    m1 : float
        Mass of the primary object (m1).
    '''
    return 0.5 * M * (1 + delta_of_eta(eta))

def m2_of_M_eta(M, eta):
    '''
    Calculate the mass of the secondary object (m2) in a binary system
    given the total mass (M) and the symmetric mass ratio (eta).

    Parameters
    ----------
    M : float
        Total mass of the binary system.
    eta : float
        Symmetric mass ratio of the binary system.

    Returns
    -------
    m2 : float
        Mass of the secondary object (m2).
    '''
    return 0.5 * M * (1 - delta_of_eta(eta))

def m1_of_Mc_eta(Mc, eta):
    """
    Calculate the component mass m1 of a binary system given the chirp mass Mc and symmetric mass ratio eta.

    Parameters
    ----------
    Mc : float
        Chirp mass of the binary system.
    eta : float
        Symmetric mass ratio of the binary system.

    Returns
    -------
    m1 : float
        Mass of the primary object (m1).
    """
    return 0.5 * M_of_Mc_eta(Mc, eta) * (1 + delta_of_eta(eta))

def m2_of_Mc_eta(Mc, eta):
    """
    Calculate the component mass m2 of a binary system given the chirp mass Mc and symmetric mass ratio eta.

    Parameters
    ----------
    Mc : float
        Chirp mass of the binary system.
    eta : float
        Symmetric mass ratio of the binary system.

    Returns
    -------
    m2 : float
        Mass of the secondary object (m2).
    """
    return 0.5 * M_of_Mc_eta(Mc, eta) * (1 - delta_of_eta(eta))

def m1_of_M_q(M, q):
    '''
    Calculate the mass of the primary object (m1) in a binary system
    given the total mass (M) and the mass ratio (q).

    Parameters
    ----------
    M : float
        Total mass of the binary system.
    q : float
        Mass ratio of the binary system.

    Returns
    -------
    m1 : float
        Mass of the primary object (m1).
    '''
    return 0.5 * M * (1 + delta_of_q(q))

def m2_of_M_q(M, q):
    '''
    Calculate the mass of the secondary object (m2) in a binary system
    given the total mass (M) and the mass ratio (q).

    Parameters
    ----------
    M : float
        Total mass of the binary system.
    q : float
        Mass ratio of the binary system.

    Returns
    -------
    m2 : float
        Mass of the secondary object (m2).
    '''
    return 0.5 * M * (1 - delta_of_q(q))

def m1_of_Mc_q(Mc, q):
    '''
    Calculate the mass of the primary object (m1) in a binary system
    given the chirp mass (Mc) and the mass ratio (q).

    Parameters
    ----------
    Mc : float
        The chirp mass of the binary system.
    q : float
        The mass ratio of the binary system.

    Returns
    -------
    m1 : float
        Mass of the primary object (m1).
    '''
    return 0.5 * M_of_Mc_q(Mc, q) * (1 + delta_of_q(q))

def m2_of_Mc_q(Mc, q):
    '''
    Calculate the mass of the secondary object (m2) in a binary system
    given the chirp mass (Mc) and the mass ratio (q).

    Parameters
    ----------
    Mc : float
        The chirp mass of the binary system.
    q : float
        The mass ratio of the binary system.

    Returns
    -------
    m2 : float
        Mass of the secondary object (m2).
    '''
    return 0.5 * M_of_Mc_q(Mc, q) * (1 - delta_of_q(q))

def m1_m2_of_M_eta(M, eta):
    '''
    Calculate the primary and secondary masses (m1, m2) of a binary system
    given the total mass M and symmetric mass ratio eta.

    Parameters
    ----------
    M : float
        Total mass.
    eta : float
        Symmetric mass ratio.

    Returms
    -------
    m1 : float
        Mass of the primary object.
    m2 : float
        Mass of the secondary object.
    '''
    delta = delta_of_eta(eta)
    return 0.5 * M * (1 + delta), 0.5 * M * (1 - delta)

def m1_m2_of_Mc_eta(Mc, eta):
    '''
    Calculate the primary and secondary masses (m1, m2) of a binary system
    given the chirp mass (Mc) and symmetric mass ratio (eta).

    Parameters
    ----------
    Mc : float
        Chirp mass of the binary system.
    eta : float
        Symmetric mass ratio of the binary system.

    Returns
    -------
    m1 : float
        Mass of the primary object.
    m2 : float
        Mass of the secondary object.
    '''
    return m1_m2_of_M_eta(M_of_Mc_eta(Mc, eta), eta)

def M_eta_of_m1_m2(m1,m2):
    '''
    Calculate the total mass (M) and symmetric mass ratio (eta) of a binary system
    given the component masses (m1, m2).

    Parameters
    ----------
    m1 : float
        Mass of the primary object.
    m2 : float
        Mass of the secondary object.

    Returns
    -------
    M : float
        Total mass of the binary system.
    eta : float
        Symmetric mass ratio of the binary system.
    '''
    return m1+m2, eta_of_q(m1/m2)

def Mc_eta_of_M_q(M, q):
    '''
    Calculate the chirp mass (Mc) and symmetric mass ratio (eta) of a binary system
    given the total mass (M) and mass ratio (q).

    Parameters
    ----------
    M : float
        Total mass of the binary system.
    q : float
        Mass ratio of the binary system.

    Returns
    -------
    Mc : float
        Chirp mass of the binary system.
    eta : float
        Symmetric mass ratio of the binary system.
    '''
    eta = eta_of_q(q)
    return Mc_of_M_eta(M, eta), eta

def Mc_eta_of_m1_m2(m1,m2):
    '''
    Calculate the chirp mass (Mc) and symmetric mass ratio (eta) of a binary system
    given the component masses (m1, m2).

    Parameters
    ----------
    m1 : float
        Mass of the primary object.
    m2 : float
        Mass of the secondary object.

    Returns
    -------
    Mc : float
        Chirp mass of the binary system.
    eta : float
        Symmetric mass ratio of the binary system.
    '''
    eta = eta_of_q(m1/m2)
    return Mc_of_M_eta(m1+m2,eta), eta


#-----spin functions-----
def chi_s(chi1,chi2):
    '''
    Calculate the symmetric spin parameter (chi_s) given the individual spins (chi1,chi2).

    Parameters
    ----------
    chi1 : float
        The spin of the primary object.
    chi2 : float
        The spin of the secondary object.

    Returns
    -------
    chi_s : float
        The symmetric spin parameter.
    '''
    return 0.5*(chi1+chi2)

def chi_a(chi1,chi2):
    '''
    Calculate the antisymmetric spin parameter (chi_a) given the individual spins (chi1,chi2).

    Parameters
    ----------
    chi1 : float
        The spin of the primary object.
    chi2 : float
        The spin of the secondary object.

    Returns
    -------
    chi_a : float
        The antisymmetric spin parameter.
    '''
    return 0.5*(chi1-chi2)

def chi_eff(m1,m2,chi1,chi2):
    '''
    Calculate the effective spin parameter (chi_eff) given the individual spins (chi1,chi2)
    and component masses (m1,m2).

    Parameters
    ----------
    m1 : float
        Mass of the primary object.
    m2 : float
        Mass of the secondary object.
    chi1 : float
        The spin of the primary object.
    chi2 : float
        The spin of the secondary object.

    Returns
    -------
    chi_eff : float
        The effective spin parameter.
    '''
    return (m1 * chi1 + m2 * chi2) / (m1+m2)


#-----distance parameter functions-----
def DL_of_z_ap_cosmo(z, cosmo=None):
    '''
    Calculate the luminosity distance (DL) given the redshift (z) using astropy.

    Parameters
    ----------
    z : float
        The redshift.
    cosmo : astropy.cosmology, optional
        The cosmology. Default is Planck18.

    Returns
    -------
    DL : float
        The luminosity distance.
    '''
    if cosmo is None: cosmo = ap.cosmology.Planck18
    return cosmo.luminosity_distance(z).value

def z_of_DL_ap_cosmo(DL, cosmo=None):
    '''
    Calculate the redshift (z) given the luminosity distance (DL) using astropy.

    Parameters
    ----------
    DL : float
        The luminosity distance.
    cosmo : astropy.cosmology, optional
        The cosmology. Default is Planck18.

    Returns
    -------
    z : float
        The redshift.
    '''
    if cosmo is None: cosmo = ap.cosmology.Planck18
    return ap.cosmology.z_at_value(cosmo.luminosity_distance, DL * ap.units.Mpc).value


#-----tidal parameter functions-----
def lam_t_of_lam_12_eta(lam1,lam2,eta): # from arXiv:1402.5156
    '''
    Calculate the effective tidal parameter (lambda_t)
    given the individual tidal parameters (lambda1, lambda2) and symmetric mass ratio (eta).

    Parameters
    ----------
    lam1 : float
        The tidal parameter of the primary object.
    lam2 : float
        The tidal parameter of the secondary object.
    eta : float
        The symmetric mass ratio of the binary system.

    Returns
    -------
    lam_t : float
        The effective tidal parameter.
    '''
    return (8./13. * ( (1. + 7. * eta - 31. * eta**2) * (lam1 + lam2) +
                       delta_of_eta(eta) * (1. + 9. * eta - 11. * eta**2) * (lam1 - lam2) ))

def delta_lam_t_of_lam_12_eta(lam1,lam2,eta): # from arXiv:1402.5156
    '''
    Calculate the delta-effective tidal parameter (delta_lambda_t)
    given the individual tidal parameters (lambda1, lambda2) and symmetric mass ratio (eta).

    Parameters
    ----------
    lam1 : float
        The tidal parameter of the primary object.
    lam2 : float
        The tidal parameter of the secondary object.
    eta : float
        The symmetric mass ratio of the binary system.

    Returns
    -------
    delta_lam_t : float
        The delta-effective tidal parameter.
    '''
    return (0.5 * ( delta_of_eta(eta) * (1319. - 13272. * eta + 8944. * eta**2) / 1319. * (lam1 + lam2)+
                          (1319. - 15910. * eta + 32850. * eta**2 + 3380. * eta**3) / 1319. * (lam1 - lam2) ))

def lam_ts_of_lam_12_eta(lam1,lam2,eta): # from arXiv:1402.5156
    '''
    Calculate the effective tidal parameter (lambda_t) and the delta-effective tidal parameter (delta_lambda_t)
    given the individual tidal parameters (lambda1, lambda2) and symmetric mass ratio (eta).

    Parameters
    ----------
    lam1 : float
        The tidal parameter of the primary object.
    lam2 : float
        The tidal parameter of the secondary object.
    eta : float
        The symmetric mass ratio of the binary system.

    Returns
    -------
    lam_t : float
        The effective tidal parameter.
    delta_lam_t : float
        The delta-effective tidal parameter.
    '''
    delta = delta_of_eta(eta)
    lam_t = 8./13. * ( (1. + 7. * eta - 31. * eta**2) * (lam1 + lam2) +
                       delta * (1. + 9. * eta - 11. * eta**2) * (lam1 - lam2) )
    delta_lam_t = 0.5 * ( delta * (1319. - 13272. * eta + 8944. * eta**2) / 1319. * (lam1 + lam2)+
                          (1319. - 15910. * eta + 32850. * eta**2 + 3380. * eta**3) / 1319. * (lam1 - lam2) )
    return lam_t, delta_lam_t

def lam1_of_lam_ts_eta(lam_t, delta_lam_t, eta):
    '''
    Calculate the tidal parameter of the primary object (lambda1)
    given the effective tidal parameter (lambda_t) and the delta-effective tidal parameter (delta_lambda_t).

    Parameters
    ----------
    lam_t : float
        The effective tidal parameter.
    delta_lam_t : float
        The delta-effective tidal parameter.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    lam1 : float
        The tidal parameter of the primary object.
    '''
    delta = delta_of_eta(eta)
    return ((-(-6.76923076923077*delta_lam_t*delta*(-0.09090909090909091 - 0.8181818181818182*eta + 1.*eta**2) +
            19.076923076923077*delta_lam_t*(-0.03225806451612903 - 0.22580645161290322*eta + 1.*eta**2) +
            3.3904473085670963*delta*(0.1474731663685152 - 1.4838998211091234*eta + 1.*eta**2)*lam_t -
            1.281273692191054*(0.39023668639053255 - 4.707100591715976*eta + 9.718934911242604*eta**2 + 1.*eta**3)*lam_t))/
            (8.881784197001252e-16*eta - 1.4210854715202004e-14*eta**2 + 2.842170943040401e-14*eta**3 + 4.500379075056848*eta**4 - 232.4912812736922*eta**5))

def lam2_of_lam_ts_eta(lam_t, delta_lam_t, eta):
    '''
    Calculate the tidal parameter of the secondary object (lambda2)
    given the effective tidal parameter (lambda_t) and the delta-effective tidal parameter (delta_lambda_t).

    Parameters
    ----------
    lam_t : float
        The effective tidal parameter.
    delta_lam_t : float
        The delta-effective tidal parameter.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    lam2 : float
        The tidal parameter of the secondary object.
    '''
    delta = delta_of_eta(eta)
    return ((delta_lam_t*(-1.5296267736621122e-19 + 3.0592535473242243e-19*delta*eta + 7.342208513578138e-18*eta**2 + 9.789611351437518e-18*eta**3 +
            (0.011550173712335778 - 9.789611351437518e-18*delta)*eta**4 + 0.11646425159938675*eta**5) +
            (-3.8240669341552804e-20 + 3.8240669341552804e-20*delta + (-4.588880320986336e-19 - 6.118507094648449e-19*delta)*eta +
            (9.789611351437518e-18 - 1.2237014189296897e-18*delta)*eta**2 + (-9.789611351437518e-18 + 1.4684417027156276e-17*delta)*eta**3 +
            (-0.0014297928149279568 - 0.007954723326344955*delta)*eta**4 + (0.07386363636363634 - 0.005511061254304498*delta)*eta**5)*lam_t)/
            (eta*(0.0909090909090909 - 0.09090909090909091*delta + (0.6363636363636364 - 0.8181818181818181*delta)*eta +
            (-2.818181818181818 + 1.*delta)*eta**2)*(-3.82026549483612e-18 + 6.112424791737792e-17*eta - 1.2224849583475584e-16*eta**2 -
            0.01935719503287065*eta**3 + 1.*eta**4)))

def lam_12_of_lam_ts_eta(lam_t,delta_lam_t,eta):
    '''
    Calculate the individual tidal parameters (lambda1, lambda2)
    given the effective tidal parameter (lambda_t) and the delta-effective tidal parameter (delta_lambda_t).

    Parameters
    ----------
    lam_t : float
        The effective tidal parameter.
    delta_lam_t : float
        The delta-effective tidal parameter.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    lam1 : float
        The tidal parameter of the primary object.
    lam2 : float
        The tidal parameter of the secondary object.
    '''
    delta = delta_of_eta(eta)
    lam1 = ((-(-6.76923076923077*delta_lam_t*delta*(-0.09090909090909091 - 0.8181818181818182*eta + 1.*eta**2) +
            19.076923076923077*delta_lam_t*(-0.03225806451612903 - 0.22580645161290322*eta + 1.*eta**2) +
            3.3904473085670963*delta*(0.1474731663685152 - 1.4838998211091234*eta + 1.*eta**2)*lam_t -
            1.281273692191054*(0.39023668639053255 - 4.707100591715976*eta + 9.718934911242604*eta**2 + 1.*eta**3)*lam_t))/
            (8.881784197001252e-16*eta - 1.4210854715202004e-14*eta**2 + 2.842170943040401e-14*eta**3 + 4.500379075056848*eta**4 - 232.4912812736922*eta**5))
    lam2 = ((delta_lam_t*(-1.5296267736621122e-19 + 3.0592535473242243e-19*delta*eta + 7.342208513578138e-18*eta**2 + 9.789611351437518e-18*eta**3 +
            (0.011550173712335778 - 9.789611351437518e-18*delta)*eta**4 + 0.11646425159938675*eta**5) +
            (-3.8240669341552804e-20 + 3.8240669341552804e-20*delta + (-4.588880320986336e-19 - 6.118507094648449e-19*delta)*eta +
            (9.789611351437518e-18 - 1.2237014189296897e-18*delta)*eta**2 + (-9.789611351437518e-18 + 1.4684417027156276e-17*delta)*eta**3 +
            (-0.0014297928149279568 - 0.007954723326344955*delta)*eta**4 + (0.07386363636363634 - 0.005511061254304498*delta)*eta**5)*lam_t)/
            (eta*(0.0909090909090909 - 0.09090909090909091*delta + (0.6363636363636364 - 0.8181818181818181*delta)*eta +
            (-2.818181818181818 + 1.*delta)*eta**2)*(-3.82026549483612e-18 + 6.112424791737792e-17*eta - 1.2224849583475584e-16*eta**2 -
            0.01935719503287065*eta**3 + 1.*eta**4)))
    return lam1, lam2


#-----derivatives of spin and mass functions-----
def del_Mc_M_of_eta(eta):
    '''
    Calculate the partial derivative of the total mass (M) with respect to the chirp mass (Mc) given the symmetric mass ratio (eta).

    Parameters
    ----------
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    float
        The partial derivative of the total mass with respect to the chirp mass.
    '''
    return eta**-0.6

def del_eta_M_of_Mc_eta(Mc,eta):
    '''
    Calculate the partial derivative of the total mass (M) with respect to the symmetric mass ratio (eta)
    given the chirp mass (Mc) and the symmetric mass ratio (eta).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    float
        The partial derivative of the symmetric mass ratio with respect to the total mass.
    '''
    return -3./5. * Mc * eta**-1.6

def del_Mc_m1_of_Mc_eta(Mc,eta):
    '''
    Calculate the partial derivative of the primary mass (m1) with respect to the chirp mass (Mc)
    given the chirp mass (Mc) and the symmetric mass ratio (eta).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    float
        The partial derivative of the primary mass with respect to the chirp mass.
    '''
    delta = delta_of_eta(eta)
    return 1./2 * del_Mc_M_of_eta(eta) * (1 + delta)

def del_eta_m1_of_Mc_eta(Mc,eta):
    '''
    Calculate the partial derivative of the primary mass (m1) with respect to the symmetric mass ratio (eta)
    given the chirp mass (Mc) and the symmetric mass ratio (eta).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    float
        The partial derivative of the primary mass with respect to the symmetric mass ratio.
    '''
    M = M_of_Mc_eta(Mc,eta)
    delta = delta_of_eta(eta)
    return 1./2 * del_eta_M_of_Mc_eta(Mc,eta) * (1 + delta) - M/delta

def del_Mc_m2_of_Mc_eta(Mc,eta):
    '''
    Calculate the partial derivative of the secondary mass (m2) with respect to the chirp mass (Mc)
    given the chirp mass (Mc) and the symmetric mass ratio (eta).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    float
        The partial derivative of the secondary mass with respect to the chirp mass.
    '''
    delta = delta_of_eta(eta)
    return 1./2 * del_Mc_M_of_eta(eta) * (1 - delta)

def del_eta_m2_of_Mc_eta(Mc,eta):
    '''
    Calculate the partial derivative of the secondary mass (m2) with respect to the symmetric mass ratio (eta)
    given the chirp mass (Mc) and the symmetric mass ratio (eta).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.

    Returns
    -------
    float
        The partial derivative of the secondary mass with respect to the symmetric mass ratio.
    '''
    M = M_of_Mc_eta(Mc,eta)
    delta = delta_of_eta(eta)
    return 1./2 * del_eta_M_of_Mc_eta(Mc,eta) * (1 - delta) + M/delta

def del_Mc_chi_eff(Mc,eta,chi1,chi2):
    '''
    Calculate the partial derivative of the effective spin parameter (chi_eff) with respect to the chirp mass (Mc)
    given the chirp mass (Mc), symmetric mass ratio (eta), and individual spins (chi1,chi2).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.
    chi1 : float
        The spin of the primary object.
    chi2 : float
        The spin of the secondary object.

    Returns
    -------
    float
        The partial derivative of the effective spin parameter with respect to the chirp mass.
    '''
    M = M_of_Mc_eta(Mc,eta)
    m1, m2 = m1_m2_of_M_eta(M,eta)
    return -1./M * del_Mc_M_of_eta(eta) * chi_eff(m1,m2,chi1,chi2) + 1./M * (del_Mc_m1_of_Mc_eta(Mc,eta) * chi1 + del_Mc_m2_of_Mc_eta(Mc,eta) * chi2)

def del_eta_chi_eff(Mc,eta,chi1,chi2):
    '''
    Calculate the partial derivative of the effective spin parameter (chi_eff) with respect to the symmetric mass ratio (eta)
    given the chirp mass (Mc), symmetric mass ratio (eta), and individual spins (chi1,chi2).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.
    chi1 : float
        The spin of the primary object.
    chi2 : float
        The spin of the secondary object.

    Returns
    -------
    float
        The partial derivative of the effective spin parameter with respect to the symmetric mass ratio.
    '''
    M = M_of_Mc_eta(Mc,eta)
    m1, m2 = m1_m2_of_M_eta(M,eta)
    return -1./M * del_eta_M_of_Mc_eta(Mc,eta) * chi_eff(m1,m2,chi1,chi2) + 1./M * (del_eta_m1_of_Mc_eta(Mc,eta) * chi1 + del_eta_m2_of_Mc_eta(Mc,eta) * chi2)

def del_chi1_chi_eff(Mc,eta,chi1,chi2):
    '''
    Calculate the partial derivative of the effective spin parameter (chi_eff) with respect to the spin of the primary object (chi1)
    given the chirp mass (Mc), symmetric mass ratio (eta), and individual spins (chi1,chi2).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.
    chi1 : float
        The spin of the primary object.
    chi2 : float
        The spin of the secondary object.

    Returns
    -------
    float
        The partial derivative of the effective spin parameter with respect to the spin of the primary object.
    '''
    M = M_of_Mc_eta(Mc,eta)
    m1, m2 = m1_m2_of_M_eta(M,eta)
    return 1./M * (m1 + m2 * chi2)

def del_chi2_chi_eff(Mc,eta,chi1,chi2):
    '''
    Calculate the partial derivative of the effective spin parameter (chi_eff) with respect to the spin of the secondary object (chi2)
    given the chirp mass (Mc), symmetric mass ratio (eta), and individual spins (chi1,chi2).

    Parameters
    ----------
    Mc : float
        The chirp mass.
    eta : float
        The symmetric mass ratio.
    chi1 : float
        The spin of the primary object.
    chi2 : float
        The spin of the secondary object.

    Returns
    -------
    float
        The partial derivative of the effective spin parameter with respect to the spin of the secondary object.
    '''
    M = M_of_Mc_eta(Mc,eta)
    m1, m2 = m1_m2_of_M_eta(M,eta)
    return 1./M * (m2 + m1 * chi1)



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np
from lal import GreenwichMeanSiderealTime

import gwbench.basic_relations as brs
import gwbench.injections as injections
import gwbench.legacy as legacy
import gwbench.network as network
import gwbench.multi_network as multi_network

np.set_printoptions(linewidth=200)

############################################################################
### User Choices
############################################################################

# injection ID
inj_id = 0

# user's choice: waveform to use
wf_model_name = 'tf2'
#wf_model_name = 'tf2_tidal'
#wf_model_name = 'lal_bbh'
#wf_model_name = 'lal_bns'

if   wf_model_name == 'tf2':       wf_other_var_dic = None
elif wf_model_name == 'tf2_tidal': wf_other_var_dic = None
elif wf_model_name == 'lal_bbh':   wf_other_var_dic = {'approximant':'IMRPhenomD'}
elif wf_model_name == 'lal_bns':   wf_other_var_dic = {'approximant':'IMRPhenomD_NRTidalv2'}

# user defined waveform model
if 1: user_waveform = None
else:
    wf_model_name    = 'tf2_user'
    wf_other_var_dic = None
    user_waveform   = {'np': '../gwbench/wf_models/tf2_np.py', 'sp':'../gwbench/wf_models/tf2_sp.py'}

# example detector location defined by user
user_locs = {'user-loc':{'longitude': 3.2, 'latitude': 0.4, 'arm_azimuth':0.3, 'which_arm':'y', 'shape':'L'}}
# example detector psd defined by the user
user_psds = {'user-tec':{'psd_file':'../gwbench/noise_curves/ce2_40km_cb.txt', 'is_asd':True}}

# user's choice: with respect to which parameters to take derivatives for the Fisher analysis
if 'tidal' in wf_model_name or 'bns' in wf_model_name: deriv_symbs_string = 'Mc eta chi1z chi2z DL tc phic iota lam_t ra dec psi'
else: deriv_symbs_string = 'Mc eta chi1z chi2z DL tc phic iota ra dec psi'

# user's choice: convert derivatives to cos or log for specific variables
conv_cos = ('dec','iota')
conv_log = ('Mc','DL','lam_t')

# user's choice: switch derivatives of DL, tc, and phic to be analytical
# TODO currently the legacy routine does not work with analytical derivatives
ana_deriv_symbs_string = None

# if symbolic derivatives, take from generate_lambdified_functions.py
# if numeric  derivatives, user's decision
use_rot = 0
# 1 for True, 0 for False
# calculate SNRs, error matrices, and errors only for the network
only_net = 1

# number of cores to use to parallelize the calculation of unique detector response derivatives
# = None for no parallelization, = 2,3,4,... to allocate N cores (even numbers preferred)
num_cores = None

# choose numdifftools parameters for numerical derivatives
step      = 1e-9
method    = 'central'
order     = 2
d_order_n = 1

# user's choice to generate injection parameters
if 'tidal' in wf_model_name or 'bns' in wf_model_name:
    mmin      = 0.8
    mmax      = 3
    chi_lo    = -0.05
    chi_hi    = 0.05
else:
    mmin      = 5
    mmax      = 100
    chi_lo    = -0.75
    chi_hi    = 0.75

cosmo_dict = {'zmin':0, 'zmax':0.2, 'sampler':'uniform_comoving_volume_inversion'}
mass_dict  = {'dist':'uniform', 'mmin':mmin, 'mmax':mmax}
# the default waveforms above are non-precessing, hence dim=1, set dim=3 for precessing waveforms like 'IMRPhenomPv2' or 'IMRPhenomPv2_NRTidalv2'
spin_dict  = {'dim':1, 'geom':'cartesian', 'chi_lo':chi_lo, 'chi_hi':chi_hi}

redshifted = 1
num_injs   = 100
seed       = 29378
file_path  = None

injections_data = injections.injections_CBC_params_redshift(cosmo_dict,mass_dict,spin_dict,redshifted,num_injs,seed,file_path)

############################################################################
### injection parameters
############################################################################

inj_params = {
    'Mc'    : injections_data[0][inj_id],
    'eta'   : injections_data[1][inj_id],
    'chi1x' : injections_data[2][inj_id],
    'chi1y' : injections_data[3][inj_id],
    'chi1z' : injections_data[4][inj_id],
    'chi2x' : injections_data[5][inj_id],
    'chi2y' : injections_data[6][inj_id],
    'chi2z' : injections_data[7][inj_id],
    'DL'    : injections_data[8][inj_id],
    'tc'    : 0.,
    'phic'  : 0.,
    'iota'  : injections_data[9][inj_id],
    'ra'    : injections_data[10][inj_id],
    'dec'   : injections_data[11][inj_id],
    'psi'   : injections_data[12][inj_id],
    'gmst0' : GreenwichMeanSiderealTime(1247227950.),
    'z'     : injections_data[13][inj_id],
    }

if 'tidal' in wf_model_name or 'bns' in wf_model_name:
    inj_params['lam_t']       = 600.
    inj_params['delta_lam_t'] = 0.

print('injections parameter: ', inj_params)
print()

############################################################################
### Network specification
############################################################################

network_specs = [ ['CE2-40-CBO_H','CE2-40-CBO_L','CE2-40-CBO_V']        ,
                  ['CE2-40-CBO_H','CE2-40-CBO_L','CE2-40-CBO_user-loc'] ,
                  ['CE2-40-CBO_H','CE2-40-CBO_V','user-tec_user-loc']   ]
print('network specs: ', network_specs)
print()

f_lo = 1.
f_hi = brs.f_isco_Msolar(brs.M_of_Mc_eta(inj_params['Mc'],inj_params['eta']))
df   = 2.**-4
f    = np.arange(f_lo,f_hi+df,df)

print('f_lo:', f_lo, '   f_hi:', f_hi, '   df:', df)
print()

############################################################################
### Precalculate the unique components common among all networks
############################################################################

unique_loc_net = legacy.unique_locs_det_responses(network_specs, f, inj_params, deriv_symbs_string,
                                                       wf_model_name, wf_other_var_dic=wf_other_var_dic,
                                                       conv_cos=conv_cos, conv_log=conv_log, use_rot=use_rot, num_cores=num_cores,
                                                       user_waveform=user_waveform, user_locs=user_locs,
                                                       ana_deriv_symbs_string=ana_deriv_symbs_string,
                                                       step=step, method=method, order=order, d_order_n=d_order_n, logger_level='INFO')

# get the unique PSDs for the various detector technologies
unique_tec_net = legacy.unique_tecs(network_specs, f, user_psds=user_psds)

############################################################################
### Perform the analysis for each network from the unique components
############################################################################

for num,network_spec in enumerate(network_specs):
    net = network.Network(network_spec, logger_level='WARNING')
    # get the correct network from the unique components calculated above
    legacy.get_det_responses_psds_from_locs_tecs(net, unique_loc_net, unique_tec_net)
    net.calc_snrs(only_net=only_net)
    net.calc_errors(only_net=only_net)
    net.calc_sky_area_90(only_net=only_net)

    print(f'network = {net.label}')
    print('snr = {net.snr}')
    print(f'errs = {net.errs}')
    print()

print('Fisher analysis done.')



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import os
import warnings
from copy import copy

import astropy.cosmology as apcosm
import numpy as np
from scipy.integrate import quad, simpson
from scipy.interpolate import interp1d
from scipy.optimize import minimize_scalar

import gwbench.basic_relations as brs
from gwbench.utils import min_max_mask

###
#-----standard values for reference-----
std_vals = [{'source':'BBH', 'dist':'power_uniform', 'mmin':5, 'mmax':100, 'alpha':-1.6},
            {'source':'BNS', 'dist':'gaussian', 'mean':1.35, 'sigma':0.15},
            {'source':'BNS', 'dist':'uniform', 'mmin':0.8, 'mmax':3}]

###
#-----CBC parameters sampler-----
def injections_CBC_params_redshift(cosmo_dict, mass_dict, spin_dict, redshifted, num_injs=10, seed=None, file_path=None):
    '''
    Generate parameters [(redshifted) masses, spins, luminosity distance, redshift, inclination, right ascension, declination, polarization angle
    for compact binary coalescence (CBC) injections.

    Parameters
    ---------
    cosmo_dict : dict
        Dictionary containing cosmological parameters.
    mass_dict : dict
        Dictionary containing mass parameters.
    spin_dict : dict
        Dictionary containing spin parameters.
    redshifted : bool
        Flag indicating whether to apply redshift to the mass parameters.
    num_injs : int
        Number of injections to generate (default: 10).
    seed : int
        Seed for the random number generator (default: None).
    file_path : str
        File path to save the generated injections (default: None).

    Returns
    -------
    Mc_vec : np.ndarray
        Chirp mass values.
    eta_vec : np.ndarray
        Symmetric mass ratio values.
    chi1x_vec : np.ndarray
        Spin x-component of the primary black hole.
    chi1y_vec : np.ndarray
        Spin y-component of the primary black hole.
    chi1z_vec : np.ndarray
        Spin z-component of the primary black hole.
    chi2x_vec : np.ndarray
        Spin x-component of the secondary black hole.
    chi2y_vec : np.ndarray
        Spin y-component of the secondary black hole.
    chi2z_vec : np.ndarray
        Spin z-component of the secondary black hole.
    DL_vec : np.ndarray
        Luminosity distance values.
    iota_vec : np.ndarray
        Inclination angle values.
    ra_vec : np.ndarray
        Right ascension values.
    dec_vec : np.ndarray
        Declination values.
    psi_vec : np.ndarray
        Polarization angle values.
    z_vec : np.ndarray
        Redshift values.
    '''
    rng   = np.random.default_rng(seed)
    seeds = rng.integers(100000, size=4)

    m1_vec, m2_vec                                                   = mass_sampler(mass_dict, num_injs, seeds[0])
    chi1x_vec, chi1y_vec, chi1z_vec, chi2x_vec, chi2y_vec, chi2z_vec = spin_sampler(spin_dict, num_injs, seeds[1])
    z_vec, DL_vec                                                    = redshift_lum_distance_sampler(cosmo_dict, num_injs, seeds[2])
    iota_vec, ra_vec, dec_vec, psi_vec                               = angle_sampler(num_injs, seeds[3])

    Mc_vec, eta_vec = get_Mc_eta(m1_vec, m2_vec)
    if redshifted: Mc_vec *= (1. + z_vec)

    params = [Mc_vec, eta_vec, chi1x_vec, chi1y_vec, chi1z_vec, chi2x_vec, chi2y_vec, chi2z_vec, DL_vec, iota_vec, ra_vec, dec_vec, psi_vec, z_vec]
    if file_path is not None: save_injections(params, file_path)
    return params

###
#-----IO functions-----
def load_injections(file_path):
    '''
    Load injection parameters from a file.

    Parameters
    ---------
    file_path : str
        File path to load the injections from.

    Returns
    -------
    params : list
        List of injection parameters (see injections.params_CBC_params_redshift for details)
    '''
    return np.transpose(np.loadtxt(file_path))

def save_injections(params,file_path):
    '''
    Save injection parameters to a file.

    Parameters
    ---------
    params : list
        List of injection parameters (see injections.params_CBC_params_redshift for details)
    file_path : str
        File path to save the injections to
    '''
    np.savetxt(os.path.join(file_path), np.transpose(np.array([params][0])), delimiter = ' ')


###
#-----angle samplers-----
def angle_sampler(num_injs, seed):
    '''
    Generate random values for the inclination, right ascension, declination, and polarization angle.
    The inclination and declination are sampled from a uniform distribution in cos(iota) and cos(dec), respectively.

    Parameters
    ----------
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator.

    Returns
    -------
    iota_vec : array
        Inclination angle values.
    ra_vec : array
        Right ascension values.
    dec_vec : array
        Declination values.
    psi_vec : array
        Polarization angle values.
    '''
    rngs = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=4)]
    iota_vec = np.arccos(rngs[0].uniform(low=-1, high=1, size=num_injs))
    ra_vec   = rngs[1].uniform(low=0., high=2.*np.pi, size=num_injs)
    dec_vec  = np.arccos(rngs[2].uniform(low=-1, high=1, size=num_injs)) - np.pi/2.
    psi_vec  = rngs[3].uniform(low=0., high=2.*np.pi, size=num_injs)
    return iota_vec, ra_vec, dec_vec, psi_vec

###
#-----spin samplers-----
def spin_sampler(spin_dict, num_injs, seed):
    '''
    Generate random values for the spins of the primary and secondary black holes.

    Parameters
    ----------
    spin_dict : dict
        Dictionary containing spin parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator.

    Returns
    -------
    chi1x_vec : np.ndarray
        Spin x-component of the primary black hole.
    chi1y_vec : np.ndarray
        Spin y-component of the primary black hole.
    chi1z_vec : np.ndarray
        Spin z-component of the primary black hole.
    chi2x_vec : np.ndarray
        Spin x-component of the secondary black hole.
    chi2y_vec : np.ndarray
        Spin y-component of the secondary black hole.
    chi2z_vec : np.ndarray
        Spin z-component of the secondary black hole.

    Note
    ----
    If the 'dist' key is not present in the spin_dict, the spins are sampled from either a 1D or 3D distribution specified
    by the 'dim' key. In the 1D case, the spins are aligned with the orbital angular momentum and the x and y components are
    set to zero while the z component is sampled from a uniform distribution in the range [chi_lo, chi_hi]. In the 3D case,
    the geometry has to be specified in the 'geom' key. If the geometry is 'cartesian', the spins are sampled from a uniform
    distribution in the range [chi_lo, chi_hi] for each component. If the geometry is 'spherical', the spins are sampled from
    a uniform distribution in the positive range [chi_lo, chi_hi] and the azimuthal and polar angles are sampled unfiromly on
    the sphere.

    If the 'dist' key is present, there are three possible distributions: 'gaussian_uniform', 'beta', and 'beta_gaussian_uniform'.
    Please check the code for the details of each distribution and additional parameters that need to be specified in the spin_dict.
    '''
    if 'dist' not in spin_dict:
        rngs   = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=6)]
        chi_lo = spin_dict['chi_lo']
        chi_hi = spin_dict['chi_hi']
        dim    = spin_dict['dim']

        if     dim == 1:
            chiz_vecs = [rngs[i].uniform(low=chi_lo, high=chi_hi, size=num_injs) for i in (2,5)]
            return [np.zeros(num_injs), np.zeros(num_injs), chiz_vecs[0], np.zeros(num_injs), np.zeros(num_injs), chiz_vecs[1]]
        elif dim == 3:
            if     spin_dict['geom'] == 'cartesian':
                return [rngs[i].uniform(low=chi_lo, high=chi_hi, size=num_injs) for i in range(6)]
            elif spin_dict['geom'] == 'spherical':
                # chi1
                chi_vec   = (rngs[0].uniform(low=chi_lo**3., high=chi_hi**3., size=num_injs))**(1./3.)
                theta_vec = np.arccos(rngs[1].uniform(low=-1, high=1, size=num_injs))
                phi_vec   = rngs[2].uniform(low=0., high=2.*np.pi, size=num_injs)
                chi1x_vec, chi1y_vec, chi1z_vec = get_cartesian_from_spherical(chi_vec,theta_vec,phi_vec)
                # chi2
                chi_vec   = (rngs[3].uniform(low=chi_lo**3., high=chi_hi**3., size=num_injs))**(1./3.)
                theta_vec = np.arccos(rngs[4].uniform(low=-1, high=1, size=num_injs))
                phi_vec   = rngs[5].uniform(low=0., high=2.*np.pi, size=num_injs)
                chi2x_vec, chi2y_vec, chi2z_vec = get_cartesian_from_spherical(chi_vec,theta_vec,phi_vec)
                return [chi1x_vec, chi1y_vec, chi1z_vec, chi2x_vec, chi2y_vec, chi2z_vec]

    elif spin_dict['dist'] == 'gaussian_uniform':
        rngs     = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=2)]
        mean     = spin_dict['mean']
        sigma    = spin_dict['sigma']
        chi1_min = spin_dict['chi1_min']
        chi1_max = spin_dict['chi1_max']
        chi2_min = spin_dict['chi2_min']
        chi2_max = spin_dict['chi2_max']

        return [np.zeros(num_injs), np.zeros(num_injs), truncated_gaussian(mean, sigma, chi1_min, chi1_max, num_injs, rng=rngs[0]),
                np.zeros(num_injs), np.zeros(num_injs), rngs[1].uniform(low=chi2_min, high=chi2_max, size=num_injs)]

    elif spin_dict['dist'] == 'beta':
        rngs  = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=2)]
        alpha = spin_dict['alpha']
        beta  = spin_dict['beta']

        return [np.zeros(num_injs), np.zeros(num_injs), rngs[0].beta(alpha, beta, num_injs),
                np.zeros(num_injs), np.zeros(num_injs), rngs[1].beta(alpha, beta, num_injs)]

    elif spin_dict['dist'] == 'beta_gaussian_uniform':
        rngs         = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=10)]
        alpha        = spin_dict['alpha']
        beta         = spin_dict['beta']
        mean         = spin_dict['mean']
        sigma        = spin_dict['sigma']
        weight       = spin_dict['weight']
        cos_min      = spin_dict['cos_min']
        cos_max      = spin_dict['cos_max']

        chi1_vec     = rngs[0].beta(alpha, beta, num_injs)
        chi2_vec     = rngs[1].beta(alpha, beta, num_injs)

        azimuth1_vec = rngs[2].uniform(low=0., high=2.*np.pi, size=num_injs)
        azimuth2_vec = rngs[3].uniform(low=0., high=2.*np.pi, size=num_injs)

        N            = int(weight * num_injs)
        M            = num_injs - N
        tilt1_vec    = np.arccos(np.concatenate((truncated_gaussian(mean, sigma, cos_min, cos_max, N, rng=rngs[4]),
                                                 rngs[5].uniform(low=cos_min, high=cos_max, size=M))))
        tilt2_vec    = np.arccos(np.concatenate((truncated_gaussian(mean, sigma, cos_min, cos_max, N, rng=rngs[6]),
                                                 rngs[7].uniform(low=cos_min, high=cos_max, size=M))))
        tilt1_vec    = tilt1_vec[rngs[8].permutation(num_injs)]
        tilt2_vec    = tilt2_vec[rngs[9].permutation(num_injs)]

        chi1xy_vec   = chi1_vec * np.sin(tilt1_vec)
        chi2xy_vec   = chi2_vec * np.sin(tilt2_vec)

        return [chi1xy_vec * np.cos(azimuth1_vec), chi1xy_vec * np.sin(azimuth1_vec), chi1_vec * np.cos(tilt1_vec),
                chi2xy_vec * np.cos(azimuth2_vec), chi2xy_vec * np.sin(azimuth2_vec), chi2_vec * np.cos(tilt2_vec)]

def get_cartesian_from_spherical(r,theta,phi):
    return r * np.sin(theta) * np.cos(phi), r * np.sin(theta) * np.sin(phi), r * np.cos(theta)


###
#-----mass samplers-----
def mass_sampler(mass_dict,num_injs,seed):
    '''
    Generate random values for the masses of the primary and secondary black holes.

    Parameters
    ----------
    mass_dict : dict
        Dictionary containing mass parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator.

    Returns
    -------
    m1_vec : np.ndarray
        Mass of the primary black hole.
    m2_vec : np.ndarray
        Mass of the secondary black hole.

    Note
    ----
    The available distributions are 'gaussian', 'double_gaussian', 'lognormal', 'power', 'power_peak', 'power_peak_uniform',
    'nsbh_power_peak_uniform', 'power_uniform', and 'uniform', specified by the 'dist' key in the mass_dict. Please refer
    to the code for the details of each distribution and additional parameters that need to be specified in the mass_dict.
    '''
    rngs = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=6)]
    if mass_dict['dist'] == 'gaussian':
        mmin  = mass_dict['mmin']
        mmax  = mass_dict['mmax']
        mean  = mass_dict['mean']
        sigma = mass_dict['sigma']
        m1_m2 = 0

        m1_vec = truncated_gaussian(mean, sigma, mmin, mmax, num_injs, rng=rngs[0])
        m2_vec = truncated_gaussian(mean, sigma, mmin, mmax, num_injs, rng=rngs[1])

    elif mass_dict['dist'] == 'double_gaussian':
        mmin   = mass_dict['mmin']
        mmax   = mass_dict['mmax']
        mean1  = mass_dict['mean1']
        sigma1 = mass_dict['sigma1']
        mean2  = mass_dict['mean2']
        sigma2 = mass_dict['sigma2']
        weight = mass_dict['weight']
        m1_m2  = 0

        N = int(weight * num_injs)
        M = num_injs - N

        m1_vec = np.concatenate((truncated_gaussian(mean1, sigma1, mmin, mmax, N, rng=rngs[0]),
                                 truncated_gaussian(mean2, sigma2, mmin, mmax, M, rng=rngs[1])))
        m2_vec = np.concatenate((truncated_gaussian(mean1, sigma1, mmin, mmax, N, rng=rngs[2]),
                                 truncated_gaussian(mean2, sigma2, mmin, mmax, M, rng=rngs[3])))
        m1_vec = m1_vec[rngs[4].permutation(num_injs)]
        m2_vec = m2_vec[rngs[5].permutation(num_injs)]

    elif mass_dict['dist'] == 'lognormal':
        mass_scale = mass_dict['mass_scale']
        mean       = mass_dict['mean']
        sigma      = mass_dict['sigma']
        m1_m2      = 1

        m1_vec = mass_scale * rngs[0].lognormal(mean, sigma, num_injs)
        m2_vec = mass_scale * rngs[1].lognormal(mean, sigma, num_injs)

    elif mass_dict['dist'] == 'power':
        mmin  = mass_dict['mmin']
        mmax  = mass_dict['mmax']
        alpha = mass_dict['alpha'] + 1
        m1_m2 = 0

        m1_vec = (mmin**alpha + (mmax**alpha - mmin**alpha)*rngs[0].random(num_injs))**(1./alpha)
        m2_vec = (mmin**alpha + (mmax**alpha - mmin**alpha)*rngs[1].random(num_injs))**(1./alpha)

    elif mass_dict['dist'] == 'power_peak':
                                                # standard power+peak parameters from GTWC-2 populations paper:
                                                # https://arxiv.org/abs/2010.14533
        mmin       = mass_dict['mmin']          # mmin       = 4.59
        mmax       = mass_dict['mmax']          # mmax       = 86.22
        m1_alpha   = mass_dict['m1_alpha']      # m1_alpha   = 2.63
        peak_frac  = mass_dict['peak_frac']     # peak_frac  = 0.1
        peak_mean  = mass_dict['peak_mean']     # peak_mu    = 33.07
        peak_sigma = mass_dict['peak_sigma']    # peak_sigma = 5.69
        delta_m    = mass_dict['delta_m']       # delta_m    = 4.82
        q_beta     = mass_dict['q_beta']        # q_beta     = 1.26
        m1_m2      = 1

        m1_vec     = power_peak(mmin, mmax, m1_alpha, peak_frac, peak_mean, peak_sigma, delta_m, num_injs, nm1s=5001, rng=rngs[0])

        q_vec      = rngs[1].power(q_beta + 1, num_injs)
        m2_vec     = q_vec * m1_vec
        m2_mask    = m2_vec < mmin

        while m2_mask.sum():
            q_vec[m2_mask] = rngs[1].power(q_beta + 1, m2_mask.sum())
            m2_vec         = q_vec * m1_vec
            m2_mask        = m2_vec < mmin

    elif mass_dict['dist'] == 'power_peak_uniform':
        mmin       = mass_dict['mmin']
        mmax       = mass_dict['mmax']
        m1_alpha   = mass_dict['m1_alpha']
        peak_frac  = mass_dict['peak_frac']
        peak_mean  = mass_dict['peak_mean']
        peak_sigma = mass_dict['peak_sigma']
        delta_m    = mass_dict['delta_m']
        m1_m2      = 1

        m1_vec     = power_peak(mmin, mmax, m1_alpha, peak_frac, peak_mean, peak_sigma, delta_m, num_injs, nm1s=5001, rng=rngs[0])
        m2_vec     = rngs[1].uniform(mmin, m1_vec)

    elif mass_dict['dist'] == 'nsbh_power_peak_uniform':
        m1_min     = mass_dict['m1_min']
        m1_max     = mass_dict['m1_max']
        m1_alpha   = mass_dict['m1_alpha']
        peak_frac  = mass_dict['peak_frac']
        peak_mean  = mass_dict['peak_mean']
        peak_sigma = mass_dict['peak_sigma']
        delta_m    = mass_dict['delta_m']
        m2_min     = mass_dict['m2_min']
        m2_max     = mass_dict['m2_max']
        m1_m2      = 1

        m1_vec     = power_peak(m1_min, m1_max, m1_alpha, peak_frac, peak_mean, peak_sigma, delta_m, num_injs, nm1s=5001, rng=rngs[0])
        m2_vec     = rngs[1].uniform(low=m2_min, high=m2_max, size=num_injs)

    elif mass_dict['dist'] == 'power_uniform':
        mmin  = mass_dict['mmin']
        mmax  = mass_dict['mmax']
        alpha = mass_dict['alpha'] + 1
        m1_m2 = 1

        m1_vec = (mmin**alpha + (mmax**alpha - mmin**alpha)*rngs[0].random(num_injs))**(1./alpha)
        m2_vec = rngs[1].uniform(mmin,m1_vec)

    elif mass_dict['dist'] == 'uniform':
        mmin  = mass_dict['mmin']
        mmax  = mass_dict['mmax']
        m1_m2 = 0

        m1_vec = rngs[0].uniform(low=mmin, high=mmax, size=num_injs)
        m2_vec = rngs[1].uniform(low=mmin, high=mmax, size=num_injs)

    elif mass_dict['dist'] == 'fixed_m1_q':
        m1    = mass_dict['m1']
        if mass_dict['q'] > 1: m2 = m1 / mass_dict['q']
        else:                  m2 = m1 * mass_dict['q']
        m1_m2 = 1

        m1_vec = m1 * np.ones(num_injs)
        m2_vec = m2 * np.ones(num_injs)

    if m1_m2: return m1_vec, m2_vec
    else:     return make_m1_m2(m1_vec,m2_vec,num_injs!=1)

#-----mass handling functions-----
def get_Mc_eta(m1_vec,m2_vec):
    '''
    Calculate the chirp mass and symmetric mass ratio from the primary and secondary black hole masses.

    Parameters
    ----------
    m1_vec : np.ndarray
        Mass of the primary black hole.
    m2_vec : np.ndarray
        Mass of the secondary black hole.

    Returns
    -------
    Mc_vec : np.ndarray
        Chirp mass values.
    eta_vec : np.ndarray
        Symmetric mass ratio values.
    '''
    eta_vec = brs.eta_of_q(m1_vec/m2_vec)
    Mc_vec  = brs.Mc_of_M_eta(m1_vec+m2_vec,eta_vec)
    return Mc_vec, eta_vec

def make_m1_m2(m1,m2,vec=True):
    '''
    Ensure that m1 >= m2.

    Parameters
    ----------
    m1 : np.ndarray
        Mass of the primary black hole.
    m2 : np.ndarray
        Mass of the secondary black hole.
    vec : bool
        Flag indicating whether to apply the operation to arrays (default: True).

    Returns
    -------
    m1 : np.ndarray
        Mass of the primary black hole.
    m2 : np.ndarray
        Mass of the secondary black hole.
    '''
    if vec:
        mt      = copy(m1)
        ids     = np.where(m1<m2)
        m1[ids] = m2[ids]
        m2[ids] = mt[ids]
        return m1, m2
    else:
        if m1 < m2: return m2, m1
        else:       return m1, m2

#-----power-peak helpers-----
def power(m, alpha):
    '''
    Power-law distribution.

    Parameters
    ----------
    m : np.ndarray
        Mass values.
    alpha : float
        Power-law exponent.

    Returns
    -------
    np.ndarray
        Power-law distribution.
    '''
    return m**(alpha)

def gaussian(m, mean, sigma):
    '''
    Gaussian distribution.

    Parameters
    ----------
    m : np.ndarray
        Mass values.
    mean : np.ndarray
        Mean value.
    sigma : np.ndarray
        Standard deviation.

    Returns
    -------
    np.ndarray
        Gaussian distribution.
    '''
    return np.exp(-((m - mean) / sigma)**2 / 2)

def smoothing(m, mmin, delta_m):
    '''
    Smoothing function.

    Parameters
    ----------
    m : np.ndarray
        Mass values.
    mmin : float
        Minimum mass value.
    delta_m : float
        Smoothing parameter.

    Returns
    -------
    np.ndarray
        Smoothed array.
    '''
    m_arr  = np.array(m)
    res    = np.zeros_like(m_arr)
    res[np.nonzero(m_arr >= mmin + delta_m)[0]] = 1
    ids    = np.nonzero(min_max_mask(m_arr, mmin, mmin + delta_m, strict_max=True))[0]
    m_arr -= mmin
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        res[ids] = 1 / (1 + np.exp(delta_m/m_arr[ids] + delta_m/(m_arr[ids] - delta_m)))
    return res

def power_peak(mmin, mmax, m1_alpha, peak_frac, peak_mean, peak_sigma, delta_m, num_injs, nm1s=5001, rng=None, seed=None):
    '''
    Power-law distribution with a Gaussian peak.

    Parameters
    ----------
    mmin : float
        Minimum mass value.
    mmax : float
        Maximum mass value.
    m1_alpha : float
        Power-law exponent.
    peak_frac : float
        Fraction of the peak.
    peak_mean : float
        Mean of the peak.
    peak_sigma : float
        Standard deviation of the peak.
    delta_m : float
        Smoothing parameter.
    num_injs : int
        Number of injections to generate.
    nm1s : int
        Number of mass values to sample (default: 5001).
    rng : np.random.Generator
        Random number generator (default: None).
    seed : int
        Seed for the random number generator (default: None).

    Returns
    -------
    np.ndarray
        Mass values.
    '''
    if rng is None: rng = np.random.default_rng(seed)
    m1s            = np.linspace(mmin, mmax, nm1s)
    power_part     = (1 - peak_frac) * power(m1s, -m1_alpha) / simpson(power(m1s, -m1_alpha), x=m1s)
    gauss_part     = peak_frac * gaussian(m1s, peak_mean, peak_sigma) / simpson(gaussian(m1s, peak_mean, peak_sigma), x=m1s)
    m1_dist        = (power_part + gauss_part) * smoothing(m1s, mmin, delta_m)
    window_cdf     = np.array([simpson(m1_dist[:i], x=m1s[:i]) for i in range(1, len(m1s)+1)]) / simpson(m1_dist, x=m1s)
    inv_window_cdf = interp1d(window_cdf, m1s)
    return inv_window_cdf(rng.random(num_injs))

#-----general helpers-----
def truncated_gaussian(mean, sigma, minv, maxv, num_injs, rng=None, seed=None):
    '''
    Truncated Gaussian distribution.

    Parameters
    ----------
    mean : float
        Mean value.
    sigma : float
        Standard deviation.
    minv : float
        Minimum value.
    maxv : float
        Maximum value.
    num_injs : int
        Number of injections to generate.
    rng : np.random.Generator
        Random number generator (default: None).
    seed : int
        Seed for the random number generator (default: None).

    Returns
    -------
    np.ndarray
        Truncated Gaussian distribution.
    '''
    if rng is None: rng = np.random.default_rng(seed)
    sample_vec = np.zeros(num_injs)
    ids        = np.arange(num_injs)
    while ids.size > 0:
        sample_vec[ids] = rng.normal(loc=mean, scale=sigma, size=ids.size)
        ids             = np.nonzero(np.logical_not(min_max_mask(sample_vec, minv, maxv, strict_min=True, strict_max=True)))[0]
    return sample_vec


###
#-----redshift and lum distance samplers-----
def redshift_lum_distance_sampler(cosmo_dict, num_injs, seed):
    '''
    Generate random values for the redshift and luminosity distance.

    Parameters
    ----------
    cosmo_dict : dict
        Dictionary containing cosmological parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator.

    Returns
    -------
    z_vec : np.ndarray
        Redshift values.
    DL_vec : np.ndarray
        Luminosity distance values.

    Note
    ----
    The available distributions are 'uniform', 'uniform_in_bins', 'uniform_comoving_volume_inversion',
    'uniform_comoving_volume_rejection', 'mdbn_rate_inversion', 'bns_md_rate_inversion', and 'popIII_rate_inversion',
    specified by the 'sampler' key in the cosmo_dict. Please refer to the code for the details of each distribution
    and additional parameters that need to be specified in the cosmo_dict.
    '''
    zmin = cosmo_dict['zmin']
    zmax = cosmo_dict['zmax']

    keys = list(cosmo_dict.keys())
    if 'Om0' in keys:  Om0 = cosmo_dict['Om0']
    else:              Om0 = None
    if 'Ode0' in keys: Ode0 = cosmo_dict['Ode0']
    else:              Ode0 = None
    if 'H0' in keys:   H0 = cosmo_dict['H0']
    else:              H0 = None

    if None in (Om0,Ode0,H0): cosmo = apcosm.Planck18
    else:                     cosmo = apcosm.LambdaCDM(H0=H0, Om0=Om0, Ode0=Ode0)

    if   cosmo_dict['sampler'] == 'uniform':
        z_vec = np.random.default_rng(seed).uniform(low=zmin, high=zmax, size=num_injs)
    elif cosmo_dict['sampler'] == 'uniform_in_bins':
        z_vec = np.concatenate([
            np.random.default_rng((i+1)*seed).uniform(low=zbin[0], high=zbin[1], size=cosmo_dict['injs_per_bin'][i])
            for i,zbin in enumerate(cosmo_dict['bins']) ])
    elif cosmo_dict['sampler'] == 'uniform_comoving_volume_inversion':
        z_vec = uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed,nzs=None)
    elif cosmo_dict['sampler'] == 'uniform_comoving_volume_rejection':
        z_vec = uniform_comoving_volume_redshift_rejection_sampler(zmin,zmax,cosmo,num_injs,seed,nzs=40)
    elif cosmo_dict['sampler'] == 'mdbn_rate_inversion':
        z_vec = mdbn_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed,nzs=None)
    elif cosmo_dict['sampler'] == 'bns_md_rate_inversion':
        z_vec = bns_md_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed,nzs=None)
    elif cosmo_dict['sampler'] == 'popIII_rate_inversion':
        z_vec = popIII_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin, zmax, cosmo, num_injs, seed, nzs=None)
    elif cosmo_dict['sampler'] == 'pbh_power_age':
        z_vec = pbh_universe_age_power_law_sampler(zmin, zmax, cosmo, num_injs, seed)

    np.random.default_rng(seed).shuffle(z_vec)

    return z_vec, cosmo.luminosity_distance(z_vec).value

#-----redshift samplers-----
def uniform_comoving_volume_redshift_inversion_sampler(zmin, zmax, cosmo, num_injs, seed=None, nzs=None):
    '''
    Generate redshift values using inversion sampling for a uniform distribution in comoving volume.

    Parameters
    ----------
    zmin : float
        Minimum redshift value.
    zmax : float
        Maximum redshift value.
    cosmo : astropy.cosmology
        Cosmological parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator (default: None).
    nzs : int
        Number of redshift bins to sample (default: None).

    Returns
    -------
    np.ndarray
        Redshift values.
    '''
    rng = np.random.default_rng(seed)
    if nzs is None: nzs = max(5001, 50 * int((zmax-zmin)) + 1)
    zs = np.linspace(zmin,zmax,nzs)
    dist = (lambda z: ((4.*np.pi*cosmo.differential_comoving_volume(z).value)/(1.+z)))(zs)
    window_cdf = np.array([simpson(dist[:i], x=zs[:i]) for i in range(1,nzs+1)]) / simpson(dist, x=zs)
    inv_window_cdf = interp1d(window_cdf, zs)
    return inv_window_cdf(rng.random(num_injs))

def uniform_comoving_volume_redshift_rejection_sampler(zmin, zmax, cosmo, num_injs, seed=None, nzs=None):
    '''
    Generate redshift values using rejection sampling for a uniform distribution in comoving volume.

    Parameters
    ----------
    zmin : float
        Minimum redshift value.
    zmax : float
        Maximum redshift value.
    cosmo : astropy.cosmology
        Cosmological parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator (default: None).
    nzs : int
        Number of redshift bins to sample (default: None).

    Returns
    -------
    np.ndarray
        Redshift values.
    '''
    rng = np.random.default_rng(seed)
    dist = lambda z: ((4.*np.pi*cosmo.differential_comoving_volume(z).value)/(1.+z))
    window_norm = quad(dist, zmin, zmax)[0]
    flip_window_pdf = lambda z: -dist(z) / window_norm
    window_pdf_max = -minimize_scalar(flip_window_pdf,bounds=[zmin,zmax],method='bounded').fun

    if nzs is None or nzs < 2: nzs = 2
    zs = np.linspace(zmin,zmax,nzs)
    segment_nums = np.asarray((num_injs * np.array([-quad(flip_window_pdf,zs[i],zs[i+1])[0] for i in range(nzs-1)])), dtype=int)
    segment_pts = np.sum(segment_nums)
    segment_maxs = np.array([-minimize_scalar(flip_window_pdf,bounds=[zs[i],zs[i+1]],method='bounded').fun for i in range(nzs-1)])

    z_sample = np.zeros(num_injs)

    for j,num in enumerate(segment_nums):
        id_shift = np.sum(segment_nums[:j+1]) - num
        ids = np.arange(num) + id_shift
        while ids.size > 0:
            z_sample[ids] = rng.uniform(zs[j],zs[j+1],ids.size)
            ids = ids[np.nonzero(rng.uniform(0.,segment_maxs[j],ids.size) >= -flip_window_pdf(z_sample[ids]))[0]]

    if nzs > 2:
        ids = np.arange(num_injs - segment_pts) + segment_pts
        while ids.size > 0:
            z_sample[ids] = rng.uniform(zmin,zmax,ids.size)
            ids = ids[np.nonzero(rng.uniform(0.,window_pdf_max,ids.size) >= -flip_window_pdf(z_sample[ids]))[0]]
        rng.shuffle(z_sample)

    return z_sample

def mdbn_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin, zmax, cosmo, num_injs, seed=None, nzs=None):
    '''
    Generate redshift values using inversion sampling for the Madau-Dickinson-Belczynski-Ng field BBH volumetric merger rate.

    Parameters
    ----------
    zmin : float
        Minimum redshift value.
    zmax : float
        Maximum redshift value.
    cosmo : astropy.cosmology
        Cosmological parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator (default: None).
    nzs : int
        Number of redshift bins to sample (default: None).

    Returns
    -------
    np.ndarray
        Redshift values.
    '''
    rng = np.random.default_rng(seed)
    if nzs is None: nzs = max(5001, 50 * int((zmax-zmin)) + 1)
    zs = np.linspace(zmin,zmax,nzs)
    dist = (lambda z: ((mdbn_merger_rate(z)*4.*np.pi*cosmo.differential_comoving_volume(z).value)/(1.+z)))(zs)
    window_cdf = np.array([simpson(dist[:i], x=zs[:i]) for i in range(1,nzs+1)]) / simpson(dist, x=zs)
    inv_window_cdf = interp1d(window_cdf, zs)
    return inv_window_cdf(rng.random(num_injs))

def bns_md_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin, zmax, cosmo, num_injs, seed=None, nzs=None):
    '''
    Generate redshift values using inversion sampling for the Madau-Dickinson star formation rate and 1/t time delay.

    Parameters
    ----------
    zmin : float
        Minimum redshift value.
    zmax : float
        Maximum redshift value.
    cosmo : astropy.cosmology
        Cosmological parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator (default: None).
    nzs : int
        Number of redshift bins to sample (default: None).

    Returns
    -------
    np.ndarray
        Redshift values.
    '''
    rng = np.random.default_rng(seed)
    if nzs is None: nzs = max(5001, 50 * int((zmax-zmin)) + 1)
    zs = np.linspace(zmin,zmax,nzs)
    dist = (lambda z: ((bns_md_merger_rate(z)*4.*np.pi*cosmo.differential_comoving_volume(z).value)/(1.+z)))(zs)
    window_cdf = np.array([simpson(dist[:i], x=zs[:i]) for i in range(1,nzs+1)]) / simpson(dist, x=zs)
    inv_window_cdf = interp1d(window_cdf, zs)
    return inv_window_cdf(rng.random(num_injs))

def popIII_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin, zmax, cosmo, num_injs, seed=None, nzs=None):
    '''
    Generate redshift values using inversion sampling for the BBH from a popIII volumetric merger rate.

    Parameters
    ----------
    zmin : float
        Minimum redshift value.
    zmax : float
        Maximum redshift value.
    cosmo : astropy.cosmology
        Cosmological parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator (default: None).
    nzs : int
        Number of redshift bins to sample (default: None).

    Returns
    -------
    np.ndarray
        Redshift values.
    '''
    rng            = np.random.default_rng(seed)
    if nzs is None: nzs = max(5001, 50 * int((zmax - zmin)) + 1)
    zs             = np.linspace(zmin, zmax, nzs)
    dist           = (lambda z: ((popIII_merger_rate(z) * 4. * np.pi * cosmo.differential_comoving_volume(z).value) / (1. + z)))(zs)
    window_cdf     = np.array([simpson(dist[:i], x=zs[:i]) for i in range(1, nzs + 1)]) / simpson(dist, x=zs)
    inv_window_cdf = interp1d(window_cdf, zs)
    return inv_window_cdf(rng.random(num_injs))

# PBH redshift distribution following a power law wrt the age of the universe (https://arxiv.org/pdf/2204.11864.pdf, Eq. (5))
def pbh_universe_age_power_law_sampler(zmin, zmax, cosmo, num_injs, seed):
    '''
    Generate redshift values for PBH using a power-law distribution with respect to the age of the universe.

    Parameters
    ----------
    zmin : float
        Minimum redshift value.
    zmax : float
        Maximum redshift value.
    cosmo : astropy.cosmology
        Cosmological parameters.
    num_injs : int
        Number of injections to generate.
    seed : int
        Seed for the random number generator (default: None).

    Returns
    -------
    np.ndarray
        Redshift values.
    '''
    rng    = np.random.default_rng(seed)
    tau_0  = cosmo.age(0)
    x_min  = cosmo.age(zmin) / tau_0
    x_max  = cosmo.age(zmax) / tau_0
    x_vec  = rng.power(-34/37 + 1, num_injs)
    x_mask = np.logical_or(x_vec < x_max, x_vec > x_min)

    while x_mask.sum():
        x_vec[x_mask] = rng.power(-34/37 + 1, x_mask.sum())
        x_mask        = np.logical_or(x_vec < x_max, x_vec > x_min)

    x_vec  *= tau_0.value
    tau_vec = np.geomspace(np.sort(x_vec)[0], np.sort(x_vec)[-1], 100)
    return np.interp(x_vec, tau_vec, apcosm.z_at_value(cosmo.age, tau_vec * tau_0.unit).value)

#-----merger rate functions-----
def mdbn_merger_rate(z, a0=2.57, b0=5.83, c0=3.36, phi0=1):
    '''
    Madau-Dickinson-Belczynski-Ng field BBH volumetric merger rate, see Eq. (C13) in https://arxiv.org/pdf/2012.09876.pdf
    with F-values from page 13 (v3 of the paper).

    Parameters
    ----------
    z : np.ndarray
        Redshift value.
    a0 : float
        Power-law exponent.
    b0 : float
        Power-law exponent.
    c0 : float
        Power-law exponent.
    phi0 : float
        Normalization factor.

    Returns
    -------
    np.ndarray
        Merger rate.
    '''
    return phi0 * (1+z)**a0 / (1+((1+z)/c0)**b0)

def bns_md_merger_rate(z, a0=1.803219571, b0=5.309821767, c0=2.837264101, phi0=8.765949529):
    '''
        Madau-Dickinson star formation rate and 1/t time delay with metalicity not taken into account.
        The fit is based on the data found in 'xtra_files/merger_rates/bns_n_dot_bns_md_merger_rate.txt'.

    Parameters
    ----------
    z : np.ndarray
        Redshift value.
    a0 : float
        Power-law exponent.
    b0 : float
        Power-law exponent.
    c0 : float
        Power-law exponent.
    phi0 : float
        Normalization factor.

    Returns
    -------
    np.ndarray
        Merger rate.
    '''
    return phi0 * (1+z)**a0 / (1+((1+z)/c0)**b0)

# BBH from popIII volumetric merger rate:
# https://arxiv.org/pdf/2012.09876.pdf, Eq. (C15) with III-values from page 13 (v3 of the paper)
def popIII_merger_rate(z, aIII=0.66, bIII=0.3, zIII=11.6):
    '''
    BBH from popIII volumetric merger rate, see Eq. (C15) in https://arxiv.org/pdf/2012.09876.pdf
    with III-values from page 13 (v3 of the paper).

    Parameters
    ----------
    z : np.ndarray
        Redshift value.
    aIII : float
        Power-law exponent.
    bIII : float
        Power-law exponent.
    zIII : float
        Redshift value.

    Returns
    -------
    np.ndarray
        Merger rate.
    '''
    return np.exp(aIII * (z - zIII)) / (bIII + aIII * np.exp((aIII + bIII) * (z - zIII)))



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from numpy import pi, euler_gamma

import gwbench.basic_relations as brs
from gwbench.utils import MTsun, strain_fac

wf_symbs_string = 'f Mc eta chi1z chi2z DL tc phic iota'

def hfpc_raw(f, Mc, eta, chi1z, chi2z, DL, tc, phic, iota, cos, sin, exp, log):
    '''
    Implementation of TaylorF2 waveform model for BBH systems that depend on the following parameters:
    - f: frequency array
    - Mc: chirp mass
    - eta: symmetric mass ratio
    - chi1x, chi1y, chi1z: dimensionless spin components of the primary BH
    - chi2x, chi2y, chi2z: dimensionless spin components of the secondary BH
    - DL: luminosity distance
    - tc: coalescence time
    - phic: coalescence phase
    - iota: inclination angle

    Parameters
    ----------
    f: np.ndarray or jnp.ndarray or sympy.Symbol
         Frequency array
    Mc: float or sympy.Symbol
         Chirp mass [solar mass]
    eta: float or sympy.Symbol
         Symmetric mass ratio
    chi1x: float or sympy.Symbol
         Dimensionless spin component of the primary BH along the x-axis
    chi1y: float or sympy.Symbol
         Dimensionless spin component of the primary BH along the y-axis
    chi1z: float or sympy.Symbol
         Dimensionless spin component of the primary BH along the z-axis
    chi2x : float or sympy.Symbol
         Dimensionless spin component of the secondary BH along the x-axis
    chi2y: float or sympy.Symbol
         Dimensionless spin component of the secondary BH along the y-axis
    chi2z: float or sympy.Symbol
         Dimensionless spin component of the secondary BH along the z-axis
    DL: float or sympy.Symbol
         Luminosity distance [Mpc]
    tc: float or sympy.Symbol
         Coalescence time [s]
    phic: float or sympy.Symbol
         Coalescence phase [rad]
    iota: float or sympy.Symbol
         Inclination angle [rad]
    cos: np.cos or jnp.cos or sympy.cos
         cosine function
    sin: np.sin or jnp.sin or sympy.sin
         sine function
    exp: np.exp or jnp.exp or sympy.exp
         exponential function
    log: np.log or jnp.log or sympy.log
         natural logarithm function

    Returns
    -------
    hfp: np.ndarray or jnp.ndarray or sympy expression
         Plus polarization waveform
    hfc: np.ndarray or jnp.ndarray or sympy expression
         Cross polarization waveform
    '''

    # Mc ... in solar mass
    # DL ... in mega parsec
    # convert to sec
    Mc    = Mc * MTsun
    DL    = DL * MTsun/strain_fac

    # get sym and asym chi combinations
    chi_s = brs.chi_s(chi1z,chi2z)
    chi_a = brs.chi_a(chi1z,chi2z)

    # Mc is in sec, e.g., Mc = 10*MTSUN_SI (for 10 solar mass)
    # DL is in sec, e.g., DL = 100*1e6*PC_SI/C_SI (for 100 Mpc)
    M     = Mc/eta**(3./5.)
    delta = (1.-4.*eta)**0.5
    v     = (pi*M*f)**(1./3.)
    flso  = brs.f_isco(M)
    vlso  = (pi*M*flso)**(1./3.)
    A     = ((5./24.)**0.5/pi**(2./3.))*(Mc**(5./6.)/DL)
    gamma = (732985./2268. - 24260.*eta/81. - 340.*eta**2/9.)*chi_s + (732985./2268. + 140.*eta/9.)*delta*chi_a

    # 3.5PN phasing (point particle limit)
    p0    = 1.

    p1    = 0

    p2    = (3715./756. + (55.*eta)/9.)

    p3    = (-16.*pi + (113.*delta*chi_a)/3. + (113./3. - (76.*eta)/3.)*chi_s)

    p4    = (15293365./508032. + (27145.*eta)/504.+ (3085.*eta**2)/72. + (-405./8. + 200.*eta)*chi_a**2 - (405.*delta*chi_a*chi_s)/4. + (-405./8. + (5.*eta)/2.)*chi_s**2)

    p5    = (38645.*pi/756. - 65.*pi*eta/9. - gamma)

    p5L   = (38645.*pi/756. - 65.*pi*eta/9. - gamma)*3*log(v/vlso)

    p6    = (11583231236531./4694215680. - 640./3.*pi**2 - 6848./21.*euler_gamma + eta*(-15737765635./3048192. + 2255./12.*pi**2) + eta*eta*76055./1728. - eta*eta*eta*127825./1296. \
            - (6848./21.)*log(4.) + pi*(2270.*delta*chi_a/3. + (2270./3. - 520.*eta)*chi_s) + (75515./144. - 8225.*eta/18.)*delta*chi_a*chi_s \
            + (75515./288. - 263245.*eta/252. - 480.*eta**2)*chi_a**2 + (75515./288. - 232415.*eta/504. + 1255.*eta**2/9.)*chi_s**2)

    p6L   = -(6848./21.)*log(v)

    p7    = (((77096675.*pi)/254016. + (378515.*pi*eta)/1512.- (74045.*pi*eta**2)/756. + (-25150083775./3048192. + (10566655595.*eta)/762048. - (1042165.*eta**2)/3024. + (5345.*eta**3)/36.
            + (14585./8. - 7270.*eta + 80.*eta**2)*chi_a**2)*chi_s + (14585./24. - (475.*eta)/6. + (100.*eta**2)/3.)*chi_s**3 + delta*((-25150083775./3048192.
            + (26804935.*eta)/6048. - (1985.*eta**2)/48.)*chi_a + (14585./24. - 2380.*eta)*chi_a**3 + (14585./8. - (215.*eta)/2.)*chi_a*chi_s**2)))

    phase = 2*pi*f*tc - phic - pi/4. + (3./(128.*v**5*eta))*(p0 + v**2*p2 + v**3*p3+ v**4*p4 + v**5*(p5+p5L) + v**6*(p6+p6L) + v**7*p7)

    hpc   = A * f**(-7./6.) * exp(1j * phase)

    return 0.5 * (1 + cos(iota)**2) * hpc, -1j * cos(iota) * hpc



from setuptools import setup, find_packages

setup(
    name='gwbench',
    version='0.8.5',
    author='Ssohrab Borhanian',
    author_email='sborhanian@gmail.com',
    description='A small tool to benchmark GW events',
    url='https://gitlab.com/sborhanian/gwbench',
    packages = find_packages(),
    package_data = {
        'gwbench.noise_curves': ['*.txt'],
                   },
    classifiers=[
        'Programming Language :: Python :: 3',
                ],
    python_requires='>=3.9',
      )
